#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2022/1/17 9:04
# @Author  : Han Yu, Li Kaiqi
# @File    : gate.py
import copy
from typing import Union
import numpy as np

from QuICT.core.utils import (
    GateType, MatrixType, SPECIAL_GATE_SET, DIAGONAL_GATE_SET, CGATE_LIST,
    PAULI_GATE_SET, CLIFFORD_GATE_SET,
    perm_decomposition, matrix_product_to_circuit
)
from QuICT.tools.exception.core import (
    TypeError, ValueError, GateAppendError, GateQubitAssignedError,
    QASMError, GateMatrixError, GateParametersAssignedError
)


class BasicGate(object):
    """ the abstract SuperClass of all basic quantum gate

    All basic quantum gate described in the framework have
    some common attributes and some common functions
    which defined in this class

    Attributes:
        name(str): the name of the gate
        controls(int): the number of the control bits of the gate
        cargs(list<int>): the list of the index of control bits in the circuit
        carg(int, read only): the first object of cargs

        targets(int): the number of the target bits of the gate
        targs(list<int>): the list of the index of target bits in the circuit
        targ(int, read only): the first object of targs

        params(list): the number of the parameter of the gate
        pargs(list): the list of the parameter
        parg(read only): the first object of pargs

        qasm_name(str, read only): gate's name in the OpenQASM 2.0
        type(GateType, read only): gate's type described by GateType

        matrix(np.array): the unitary matrix of the quantum gate act on targets
    """
    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def matrix(self) -> np.ndarray:
        return self._matrix

    @matrix.setter
    def matrix(self, matrix) -> np.ndarray:
        self._matrix = matrix

    @property
    def target_matrix(self) -> np.ndarray:
        return self.matrix

    @property
    def type(self):
        return self._type

    @property
    def matrix_type(self):
        return self._matrix_type

    @property
    def controls(self) -> int:
        return self._controls

    @controls.setter
    def controls(self, controls: int):
        assert isinstance(controls, int), TypeError("BasicGate.controls", "int", type(controls))
        self._controls = controls

    @property
    def cargs(self):
        return self._cargs

    @cargs.setter
    def cargs(self, cargs: Union[list, int]):
        if isinstance(cargs, int):
            cargs = [cargs]

        assert len(cargs) == len(set(cargs)), ValueError("BasicGate.cargs", "not have duplicated value", cargs)
        self._cargs = cargs

    @property
    def targets(self) -> int:
        return self._targets

    @targets.setter
    def targets(self, targets: int):
        assert isinstance(targets, int), TypeError("BasicGate.targets", "int", type(targets))
        self._targets = targets

    @property
    def targs(self):
        return self._targs

    @targs.setter
    def targs(self, targs: list):
        if isinstance(targs, int):
            targs = [targs]

        assert len(targs) == len(set(targs)), ValueError("BasicGate.targs", "not have duplicated value", targs)
        assert not set(self._cargs) & set(targs), ValueError(
            "BasicGate.targs", "have no same index with control qubits", set(self._cargs) & set(targs)
        )
        self._targs = targs

    @property
    def params(self) -> int:
        return self._params

    @params.setter
    def params(self, params: int):
        self._params = params

    @property
    def pargs(self):
        return self._pargs

    @pargs.setter
    def pargs(self, pargs: list):
        if isinstance(pargs, list):
            self._pargs = pargs
        else:
            self._pargs = [pargs]

        if len(self._pargs) != self.params:
            raise ValueError("BasicGate.pargs:length", f"equal to gate's parameter number {self._pargs}", len(pargs))

    @property
    def parg(self):
        return self.pargs[0]

    @property
    def carg(self):
        return self.cargs[0]

    @property
    def targ(self):
        return self.targs[0]

    @property
    def precision(self):
        return self._precision

    @property
    def qasm_name(self):
        return self._qasm_name

    def __init__(
        self,
        controls: int,
        targets: int,
        params: int,
        type_: GateType,
        matrix_type: MatrixType = MatrixType.normal
    ):
        self._matrix = None

        self._controls = controls
        self._targets = targets
        self._params = params
        self._cargs = []    # list of int
        self._targs = []    # list of int
        self._pargs = []    # list of float/..

        assert isinstance(type_, GateType), TypeError("BasicGate.type", "GateType", type(type_))
        self._type = type_
        self._matrix_type = matrix_type
        self._precision = np.complex128
        self._qasm_name = str(type_.name)
        self._name = "-".join([str(type_), "", ""])

        self.assigned_qubits = []   # list of qubits

    def __or__(self, targets):
        """deal the operator '|'

        Use the syntax "gate | circuit" or "gate | Composite Gate"
        to add the gate into the circuit or composite gate
        Some Examples are like this:

        X       | circuit
        CX      | circuit([0, 1])
        Measure | CompositeGate

        Note that the order of qubits is that control bits first
        and target bits followed.

        Args:
            targets: the targets the gate acts on, it can have following form,
                1) Circuit
                2) CompositeGate

        Raise:
            TypeError: the type of other is wrong
        """
        try:
            targets.append(self)
        except Exception as e:
            raise GateAppendError(f"Failure to append gate {self.name} to targets, due to {e}")

    def __and__(self, targets):
        """deal the operator '&'

        Use the syntax "gate & int" or "gate & list<int>" to set gate's attribute.
        Special uses when in composite gate's context.

        Some Examples are like this:
        X       & 1
        CX      & [0, 1]

        Note that the order of qubits is that control bits first
        and target bits followed.

        Args:
            targets: the targets the gate acts on, it can have following form,
                1) int
                2) list<int>

        Raise:
            TypeError: the type of targets is wrong
        """
        _gate = self.copy()

        if isinstance(targets, int):
            assert _gate.is_single(), GateQubitAssignedError("The qubits number should equal to the quantum gate.")

            _gate.targs = [targets]
        elif isinstance(targets, list):
            if len(targets) != _gate.controls + _gate.targets:
                raise GateQubitAssignedError("The qubits number should equal to the quantum gate.")

            _gate.cargs = targets[:_gate.controls]
            _gate.targs = targets[_gate.controls:]
        else:
            raise TypeError("BasicGate.&", "int or list<int>", type(targets))

        if CGATE_LIST:
            CGATE_LIST[-1].append(_gate)
        else:
            return _gate

    def __call__(self):
        """ give parameters for the gate, and give parameters by "()", and parameters should be one of int/float/complex

        Some Examples are like this:

        Rz(np.pi / 2)           | qubit
        U3(np.pi / 2, 0, 0)     | qubit

        *Important*: There is no parameters for current quantum gate.

        Returns:
            BasicGate: the gate after filled by parameters
        """
        return self.copy()

    def __eq__(self, other):
        assert isinstance(other, BasicGate), TypeError("BasicGate.==", "BasicGate", type(other))
        if (
            self.type != other.type or
            (self.cargs + self.targs) != (other.cargs + other.targs) or
            not np.allclose(self.matrix, other.matrix)
        ):
            return False

        return True

    def update_name(self, qubit_id: str, circuit_idx: int = None):
        """ Updated gate's name with the given information

        Args:
            qubit_id (str): The qubit's unique ID.
            circuit_idx (int, optional): The gate's order index in the circuit. Defaults to None.
        """
        qubit_id = qubit_id[:6]
        name_parts = self.name.split('-')
        name_parts[1] = qubit_id

        if circuit_idx is not None:
            name_parts[2] = str(circuit_idx)

        self.name = '-'.join(name_parts)

    def __str__(self):
        """ get gate information """
        gate_info = {
            "name": self.name,
            "controls": self.controls,
            "control_bit": self.cargs,
            "targets": self.targets,
            "target_bit": self.targs,
            "parameters": self.pargs
        }

        return str(gate_info)

    def qasm(self):
        """ generator OpenQASM string for the gate

        Return:
            string: the OpenQASM 2.0 describe of the gate
        """
        if self.type in SPECIAL_GATE_SET[4:]:
            raise QASMError(f"The gate do not support qasm, {self.type}")

        qasm_string = self.qasm_name
        if self.params > 0:
            params = [str(parg) for parg in self.pargs]
            params_string = "(" + ", ".join(params) + ")"

            qasm_string += params_string

        ctargs = [f"q[{ctarg}]" for ctarg in self.cargs + self.targs]
        ctargs_string = " " + ', '.join(ctargs) + ";\n"
        qasm_string += ctargs_string

        return qasm_string

    def convert_precision(self):
        """ Convert gate's precision into single precision np.complex64. """
        if self.type in [GateType.measure, GateType.reset, GateType.barrier]:
            return

        self._precision = np.complex64 if self._precision == np.complex128 else np.complex128
        if self.params == 0:
            self._matrix = self.matrix.astype(self._precision)

    def inverse(self):
        """ the inverse of the gate

        Return:
            BasicGate: the inverse of the gate
        """
        return self.copy()

    def commutative(self, goal, eps=1e-7):
        """ decide whether gate is commutative with another gate

        note when the gate is special gates like Unitary, Permutation, Measure and so on, return False.

        Args:
            goal(BasicGate): the target gate
            eps(float): the precision of comparision

        Return:
            bool: True if commutative
        """
        # Check the affected qubits
        self_controls = set(self.cargs)
        self_targets = set(self.targs)
        goal_controls = set(goal.cargs)
        goal_targets = set(goal.targs)
        # If the affected qubits of the gates are completely different, they must commute
        self_qubits = self_controls | self_targets
        goal_qubits = goal_controls | goal_targets
        if len(self_qubits & goal_qubits) == 0:
            return True

        # Ignore all the special gates except for the unitary gates due to the difficulty of getting the matrices
        if (
            (self.is_special() and self.type != GateType.unitary) or
            (goal.is_special() and self.type != GateType.unitary)
        ):
            return False

        # Check the target matrices of the gates
        A = self.target_matrix
        B = goal.target_matrix
        # It means commuting that any of the target matrices is close to identity
        if (
            np.allclose(A, np.identity(1 << self.targets), rtol=eps, atol=eps) or
            np.allclose(B, np.identity(1 << goal.targets), rtol=eps, atol=eps)
        ):
            return True

        # For gates whose number of target qubits is 1, optimized judgment could be used
        if self.targets == 1 and goal.targets == 1:
            # Diagonal target gates commutes with the control qubits
            if (
                (len(self_controls & goal_targets) > 0 and not goal.is_diagonal()) or
                (len(goal_controls & self_targets) > 0 and not self.is_diagonal())
            ):
                return False
            # Compute the target matrix commutation
            if (
                len(goal_targets & self_targets) > 0 and
                not np.allclose(A.dot(B), B.dot(A), rtol=eps, atol=eps)
            ):
                return False

            return True
        # Otherwise, we need to calculate the matrix commutation directly
        else:
            # Collect all the affected qubits and create the converter to the minimal qubits
            qubits = self_qubits | goal_qubits
            qubits_dict = {}
            for idx, x in enumerate(qubits):
                qubits_dict[x] = idx
            # Create two masked gates whose affected qubits are the minimal ones
            self_masked = self.cargs + self.targs
            for i in range(len(self_masked)):
                self_masked[i] = qubits_dict[self_masked[i]]
            goal_masked = goal.cargs + goal.targs
            for i in range(len(goal_masked)):
                goal_masked[i] = qubits_dict[goal_masked[i]]
            # Compute the matrix commutation
            self_matrix = self.expand(list(qubits))
            goal_matrix = goal.expand(list(qubits))
            return np.allclose(self_matrix.dot(goal_matrix), goal_matrix.dot(self_matrix), rtol=eps, atol=eps)

    def is_single(self) -> bool:
        """ judge whether gate is a one qubit gate(excluding special gate like measure, reset, custom and so on)

        Returns:
            bool: True if it is a one qubit gate
        """
        return self.targets + self.controls == 1

    def is_control_single(self) -> bool:
        """ judge whether gate has one control bit and one target bit

        Returns:
            bool: True if it is has one control bit and one target bit
        """
        return self.controls == 1 and self.targets == 1

    def is_clifford(self) -> bool:
        """ judge whether gate's matrix is a Clifford gate

        Returns:
            bool: True if gate's matrix is a Clifford gate
        """
        return self.type in CLIFFORD_GATE_SET

    def is_diagonal(self) -> bool:
        """ judge whether gate's matrix is diagonal

        Returns:
            bool: True if gate's matrix is diagonal
        """
        return (
            self.type in DIAGONAL_GATE_SET or
            (self.type == GateType.unitary and self._is_diagonal())
        )

    def _is_diagonal(self) -> bool:
        return np.allclose(np.diag(np.diag(self.matrix)), self.matrix)

    def is_pauli(self) -> bool:
        """ judge whether gate's matrix is a Pauli gate

        Returns:
            bool: True if gate's matrix is a Pauli gate
        """
        return self.type in PAULI_GATE_SET

    def is_special(self) -> bool:
        """ judge whether gate's is special gate, which is one of
        [Measure, Reset, Barrier, Perm, Unitary, ...]

        Returns:
            bool: True if gate's matrix is special
        """
        return self.type in SPECIAL_GATE_SET

    def is_identity(self) -> bool:
        if self.type in [GateType.reset, GateType.measure, GateType.barrier]:
            return False

        return np.allclose(self.matrix, np.identity(1 << (self.controls + self.targets), dtype=self.precision))

    def expand(self, qubits: Union[int, list]) -> bool:
        """ expand self matrix into the circuit's unitary linear space. If input qubits is integer, please make sure
        the indexes of current gate is within [0, qubits).

        Args:
            qubits Union[int, list]: the total number of qubits of the target circuit or the indexes of expand qubits.
        """
        if isinstance(qubits, int):
            qubits = list(range(qubits))

        qubits_num = len(qubits)
        if qubits_num == self.controls + self.targets:
            return self.matrix

        assert qubits_num > self.controls + self.targets, GateQubitAssignedError(
            "The expand qubits' num should >= gate's qubits num."
        )
        gate_args = self.cargs + self.targs
        if len(gate_args) == 0:     # Deal with not assigned quantum gate
            gate_args = [qubits[i] for i in range(self.controls + self.targets)]

        updated_args = [qubits.index(garg) for garg in gate_args]
        return matrix_product_to_circuit(self.matrix, updated_args, qubits_num)

    def copy(self):
        """ return a copy of this gate

        Returns:
            gate(BasicGate): a copy of this gate
        """
        class_name = str(self.__class__.__name__)
        gate = globals()[class_name]()

        if gate.type in SPECIAL_GATE_SET:
            gate.controls = self.controls
            gate.targets = self.targets
            gate.params = self.params

        if gate.type in [GateType.perm, GateType.unitary, GateType.perm_fx]:
            gate.matrix = self.matrix

        gate.pargs = copy.deepcopy(self.pargs)
        gate.targs = copy.deepcopy(self.targs)
        gate.cargs = copy.deepcopy(self.cargs)

        if self.assigned_qubits:
            gate.assigned_qubits = copy.deepcopy(self.assigned_qubits)
            gate.update_name(gate.assigned_qubits[0].id)

        if self.precision == np.complex64:
            gate.convert_precision()

        return gate

    def permit_element(self, element):
        """ judge whether the type of a parameter is int/float/complex

        for a quantum gate, the parameter should be int/float/complex

        Args:
            element: the element to be judged

        Returns:
            bool: True if the type of element is int/float/complex
        """
        if not isinstance(element, (int, float, complex)):
            tp = type(element)
            if tp == np.int64 or tp == np.float64 or tp == np.complex128:
                return True

            raise TypeError(self.type, "int/float/complex", type(element))


class HGate(BasicGate):
    """ Hadamard gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.h
        )

        self.matrix = np.array([
            [1 / np.sqrt(2), 1 / np.sqrt(2)],
            [1 / np.sqrt(2), -1 / np.sqrt(2)]
        ], dtype=self._precision)


H = HGate()


class HYGate(BasicGate):
    """ Self-inverse gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.hy
        )

        self.matrix = np.array([
            [1 / np.sqrt(2), -1j / np.sqrt(2)],
            [1j / np.sqrt(2), -1 / np.sqrt(2)]
        ], dtype=self._precision)


Hy = HYGate()


class SGate(BasicGate):
    """ S gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.s,
            matrix_type=MatrixType.control
        )

        self.matrix = np.array([
            [1, 0],
            [0, 1j]
        ], dtype=self._precision)

    def inverse(self):
        """ change it be sdg gate"""
        _Sdagger = SDaggerGate()
        _Sdagger.targs = copy.deepcopy(self.targs)
        _Sdagger.assigned_qubits = copy.deepcopy(self.assigned_qubits)

        return _Sdagger


S = SGate()


class SDaggerGate(BasicGate):
    """ The conjugate transpose of Phase gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.sdg,
            matrix_type=MatrixType.control
        )

        self.matrix = np.array([
            [1, 0],
            [0, -1j]
        ], dtype=self._precision)

    def inverse(self):
        """ change it to be s gate """
        _Sgate = SGate()
        _Sgate.targs = copy.deepcopy(self.targs)
        _Sgate.assigned_qubits = copy.deepcopy(self.assigned_qubits)

        return _Sgate


S_dagger = SDaggerGate()


class XGate(BasicGate):
    """ Pauli-X gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.x,
            matrix_type=MatrixType.swap
        )

        self.matrix = np.array([
            [0, 1],
            [1, 0]
        ], dtype=self._precision)


X = XGate()


class YGate(BasicGate):
    """ Pauli-Y gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.y,
            matrix_type=MatrixType.reverse
        )

        self.matrix = np.array([
            [0, -1j],
            [1j, 0]
        ], dtype=self._precision)


Y = YGate()


class ZGate(BasicGate):
    """ Pauli-Z gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.z,
            matrix_type=MatrixType.control
        )

        self.matrix = np.array([
            [1, 0],
            [0, -1]
        ], dtype=self._precision)


Z = ZGate()


class SXGate(BasicGate):
    """ sqrt(X) gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.sx
        )

        self.matrix = np.array([
            [0.5 + 0.5j, 0.5 - 0.5j],
            [0.5 - 0.5j, 0.5 + 0.5j]
        ], dtype=self._precision)

    def inverse(self):
        """ change it be rx gate"""
        _Rx = RxGate([-np.pi / 2])
        _Rx.targs = copy.deepcopy(self.targs)

        return _Rx


SX = SXGate()


class SYGate(BasicGate):
    """ sqrt(Y) gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.sy
        )

        self.matrix = np.array([
            [1 / np.sqrt(2), -1 / np.sqrt(2)],
            [1 / np.sqrt(2), 1 / np.sqrt(2)]
        ], dtype=self._precision)

    def inverse(self):
        """ change it to be ry gate"""
        _Ry = RyGate([-np.pi / 2])
        _Ry.targs = copy.deepcopy(self.targs)

        return _Ry


SY = SYGate()


class SWGate(BasicGate):
    """ sqrt(W) gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.sw
        )

        self.matrix = np.array([
            [1 / np.sqrt(2), -np.sqrt(1j / 2)],
            [np.sqrt(-1j / 2), 1 / np.sqrt(2)]
        ], dtype=self._precision)

    def inverse(self):
        """ change it be U2 gate"""
        _U2 = U2Gate([3 * np.pi / 4, 5 * np.pi / 4])
        _U2.targs = copy.deepcopy(self.targs)

        return _U2


SW = SWGate()


class IDGate(BasicGate):
    """ Identity gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.id,
            matrix_type=MatrixType.diagonal
        )

        self.matrix = np.array([
            [1, 0],
            [0, 1]
        ], dtype=self._precision)


ID = IDGate()


class U1Gate(BasicGate):
    """ Diagonal single-qubit gate """
    def __init__(self, params: list = [np.pi / 2]):
        super().__init__(
            controls=0,
            targets=1,
            params=1,
            type_=GateType.u1,
            matrix_type=MatrixType.control
        )

        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return U1Gate([alpha])

    @property
    def matrix(self):
        return np.array([
            [1, 0],
            [0, np.exp(1j * self.pargs[0])]
        ], dtype=self._precision)

    def inverse(self):
        _U1 = self.copy()
        _U1.pargs = [-self.pargs[0]]

        return _U1


U1 = U1Gate()


class U2Gate(BasicGate):
    """ One-pulse single-qubit gate """
    def __init__(self, params: list = [np.pi / 2, np.pi / 2]):
        super().__init__(
            controls=0,
            targets=1,
            params=2,
            type_=GateType.u2
        )

        self.pargs = params

    def __call__(self, alpha, beta):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate
            beta (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        params = [alpha, beta]

        for param in params:
            self.permit_element(param)

        return U2Gate(params)

    @property
    def matrix(self):
        sqrt2 = 1 / np.sqrt(2)
        return np.array([
            [1 * sqrt2,
             -np.exp(1j * self.pargs[1]) * sqrt2],
            [np.exp(1j * self.pargs[0]) * sqrt2,
             np.exp(1j * (self.pargs[0] + self.pargs[1])) * sqrt2]
        ], dtype=self._precision)

    def inverse(self):
        _U2 = self.copy()
        _U2.pargs = [np.pi - self.pargs[1], np.pi - self.pargs[0]]

        return _U2


U2 = U2Gate()


class U3Gate(BasicGate):
    """ Two-pulse single-qubit gate """
    def __init__(self, params: list = [0, 0, np.pi / 2]):
        super().__init__(
            controls=0,
            targets=1,
            params=3,
            type_=GateType.u3
        )

        self.pargs = params

    def __call__(self, alpha, beta, gamma):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate
            beta (int/float/complex): The parameter for gate
            gamma (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        params = [alpha, beta, gamma]

        for param in params:
            self.permit_element(param)

        return U3Gate(params)

    @property
    def matrix(self):
        return np.array([
            [np.cos(self.pargs[0] / 2),
             -np.exp(1j * self.pargs[2]) * np.sin(self.pargs[0] / 2)],
            [np.exp(1j * self.pargs[1]) * np.sin(self.pargs[0] / 2),
             np.exp(1j * (self.pargs[1] + self.pargs[2])) * np.cos(self.pargs[0] / 2)]
        ], dtype=self._precision)

    def inverse(self):
        _U3 = self.copy()
        _U3.pargs = [self.pargs[0], np.pi - self.pargs[2], np.pi - self.pargs[1]]

        return _U3


U3 = U3Gate()


class RxGate(BasicGate):
    """ Rotation around the x-axis gate """
    def __init__(self, params: list = [np.pi / 2]):
        super().__init__(
            controls=0,
            targets=1,
            params=1,
            type_=GateType.rx
        )

        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return RxGate([alpha])

    @property
    def matrix(self):
        return np.array([
            [np.cos(self.parg / 2), 1j * -np.sin(self.parg / 2)],
            [1j * -np.sin(self.parg / 2), np.cos(self.parg / 2)]
        ], dtype=self._precision)

    def inverse(self):
        _Rx = self.copy()
        _Rx.pargs = [-self.pargs[0]]

        return _Rx


Rx = RxGate()


class RyGate(BasicGate):
    """ Rotation around the y-axis gate """
    def __init__(self, params: list = [np.pi / 2]):
        super().__init__(
            controls=0,
            targets=1,
            params=1,
            type_=GateType.ry
        )

        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return RyGate([alpha])

    @property
    def matrix(self):
        return np.array([
            [np.cos(self.pargs[0] / 2), -np.sin(self.pargs[0] / 2)],
            [np.sin(self.pargs[0] / 2), np.cos(self.pargs[0] / 2)],
        ], dtype=self._precision)

    def inverse(self):
        _Ry = self.copy()
        _Ry.pargs = [-self.pargs[0]]

        return _Ry


Ry = RyGate()


class RzGate(BasicGate):
    """ Rotation around the z-axis gate """
    def __init__(self, params: list = [np.pi / 2]):
        super().__init__(
            controls=0,
            targets=1,
            params=1,
            type_=GateType.rz,
            matrix_type=MatrixType.diagonal
        )

        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return RzGate([alpha])

    @property
    def matrix(self):
        return np.array([
            [np.exp(-self.parg / 2 * 1j), 0],
            [0, np.exp(self.parg / 2 * 1j)]
        ], dtype=self._precision)

    def inverse(self):
        _Rz = self.copy()
        _Rz.pargs = [-self.pargs[0]]

        return _Rz


Rz = RzGate()


class TGate(BasicGate):
    """ T gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.t,
            matrix_type=MatrixType.control
        )

        self.matrix = np.array([
            [1, 0],
            [0, 1 / np.sqrt(2) + 1j * 1 / np.sqrt(2)]
        ], dtype=self._precision)

    def inverse(self):
        """ change it be tdg gate"""
        _Tdagger = TDaggerGate()
        _Tdagger.targs = copy.deepcopy(self.targs)
        _Tdagger.assigned_qubits = copy.deepcopy(self.assigned_qubits)

        return _Tdagger


T = TGate()


class TDaggerGate(BasicGate):
    """ The conjugate transpose of T gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.tdg,
            matrix_type=MatrixType.control
        )

        self.matrix = np.array([
            [1, 0],
            [0, 1 / np.sqrt(2) + 1j * -1 / np.sqrt(2)]
        ], dtype=self._precision)

    def inverse(self):
        """ change it to be t gate """
        _Tgate = TGate()
        _Tgate.targs = copy.deepcopy(self.targs)
        _Tgate.assigned_qubits = copy.deepcopy(self.assigned_qubits)

        return _Tgate


T_dagger = TDaggerGate()


class PhaseGate(BasicGate):
    """ Phase gate """
    def __init__(self, params: list = [0]):
        super().__init__(
            controls=0,
            targets=1,
            params=1,
            type_=GateType.phase,
            matrix_type=MatrixType.control
        )

        self.pargs = params
        self._qasm_name = "p"

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return PhaseGate([alpha])

    @property
    def matrix(self):
        return np.array([
            [1, 0],
            [0, np.exp(self.parg * 1j)]
        ], dtype=self._precision)

    def inverse(self):
        _Phase = self.copy()
        _Phase.pargs = [-self.parg]

        return _Phase


Phase = PhaseGate()


class GlobalPhaseGate(BasicGate):
    """ Phase gate """
    def __init__(self, params: list = [0]):
        super().__init__(
            controls=0,
            targets=1,
            params=1,
            type_=GateType.gphase,
            matrix_type=MatrixType.diagonal
        )
        self._qasm_name = "phase"
        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return GlobalPhaseGate([alpha])

    @property
    def matrix(self):
        return np.array([
            [np.exp(self.parg * 1j), 0],
            [0, np.exp(self.parg * 1j)]
        ], dtype=self._precision)

    def inverse(self):
        _Phase = self.copy()
        _Phase.pargs = [-self.parg]

        return _Phase


GPhase = GlobalPhaseGate()


class CZGate(BasicGate):
    """ controlled-Z gate """
    def __init__(self):
        super().__init__(
            controls=1,
            targets=1,
            params=0,
            type_=GateType.cz,
            matrix_type=MatrixType.control
        )

        self.matrix = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, -1]
        ], dtype=self._precision)

        self._target_matrix = np.array([
            [1, 0],
            [0, -1]
        ], dtype=self._precision)

    @property
    def target_matrix(self):
        return self._target_matrix


CZ = CZGate()


class CXGate(BasicGate):
    """ controlled-X gate """
    def __init__(self):
        super().__init__(
            controls=1,
            targets=1,
            params=0,
            type_=GateType.cx,
            matrix_type=MatrixType.reverse
        )

        self.matrix = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 0, 1],
            [0, 0, 1, 0]
        ], dtype=self._precision)

        self._target_matrix = np.array([
            [0, 1],
            [1, 0]
        ], dtype=self._precision)

    @property
    def target_matrix(self):
        return self._target_matrix


CX = CXGate()


class CYGate(BasicGate):
    """ controlled-Y gate """
    def __init__(self):
        super().__init__(
            controls=1,
            targets=1,
            params=0,
            type_=GateType.cy,
            matrix_type=MatrixType.reverse
        )

        self.matrix = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 0, -1j],
            [0, 0, 1j, 0]
        ], dtype=self._precision)

        self._target_matrix = np.array([
            [0, -1j],
            [1j, 0]
        ], dtype=self._precision)

    @property
    def target_matrix(self):
        return self._target_matrix


CY = CYGate()


class CHGate(BasicGate):
    """ controlled-Hadamard gate """
    def __init__(self):
        super().__init__(
            controls=1,
            targets=1,
            params=0,
            type_=GateType.ch
        )

        self.matrix = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1 / np.sqrt(2), 1 / np.sqrt(2)],
            [0, 0, 1 / np.sqrt(2), -1 / np.sqrt(2)]
        ], dtype=self._precision)

        self._target_matrix = np.array([
            [1 / np.sqrt(2), 1 / np.sqrt(2)],
            [1 / np.sqrt(2), -1 / np.sqrt(2)]
        ], dtype=self._precision)

    @property
    def target_matrix(self):
        return self._target_matrix


CH = CHGate()


class CRzGate(BasicGate):
    """ controlled-Rz gate """

    def __init__(self, params: list = [np.pi / 2]):
        super().__init__(
            controls=1,
            targets=1,
            params=1,
            type_=GateType.crz,
            matrix_type=MatrixType.diagonal
        )

        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return CRzGate([alpha])

    @property
    def matrix(self):
        return np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, np.exp(-self.parg / 2 * 1j), 0],
            [0, 0, 0, np.exp(self.parg / 2 * 1j)]
        ], dtype=self._precision)

    @property
    def target_matrix(self) -> np.ndarray:
        return np.array([
            [np.exp(-self.parg / 2 * 1j), 0],
            [0, np.exp(self.parg / 2 * 1j)]
        ], dtype=self._precision)

    def inverse(self):
        _CRz = self.copy()
        _CRz.pargs = [-self.pargs[0]]

        return _CRz


CRz = CRzGate()


class CU1Gate(BasicGate):
    """ Controlled-U1 gate """
    def __init__(self, params: list = [np.pi / 2]):
        super().__init__(
            controls=1,
            targets=1,
            params=1,
            type_=GateType.cu1,
            matrix_type=MatrixType.control
        )

        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return CU1Gate([alpha])

    @property
    def matrix(self):
        return np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, np.exp(1j * self.pargs[0])]
        ], dtype=self._precision)

    @property
    def target_matrix(self) -> np.ndarray:
        return np.array([
            [1, 0],
            [0, np.exp(1j * self.pargs[0])]
        ], dtype=self._precision)

    def inverse(self):
        _CU1 = self.copy()
        _CU1.pargs = [-self.pargs[0]]

        return _CU1

    def build_gate(self):
        from QuICT.core.gate import CompositeGate

        cgate = CompositeGate()
        with cgate:
            CRz(self.parg) & [0, 1]
            U1(self.parg / 2) & 0

        args = self.cargs + self.targs
        if len(args) == self.controls + self.targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


CU1 = CU1Gate()


class CU3Gate(BasicGate):
    """ Controlled-U3 gate """
    def __init__(self, params: list = [np.pi / 2, 0, 0]):
        super().__init__(
            controls=1,
            targets=1,
            params=3,
            type_=GateType.cu3
        )

        self.pargs = params

    def __call__(self, alpha, beta, gamma):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate
            beta (int/float/complex): The parameter for gate
            gamma (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        params = [alpha, beta, gamma]

        for param in params:
            self.permit_element(param)

        return CU3Gate(params)

    @property
    def matrix(self):
        return np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, np.cos(self.pargs[0] / 2), -np.exp(1j * self.pargs[2]) * np.sin(self.pargs[0] / 2)],
            [0, 0, np.exp(1j * self.pargs[1]) * np.sin(self.pargs[0] / 2),
             np.exp(1j * (self.pargs[1] + self.pargs[2])) * np.cos(self.pargs[0] / 2)]
        ], dtype=self._precision)

    @property
    def target_matrix(self) -> np.ndarray:
        return np.array([
            [np.cos(self.pargs[0] / 2), -np.exp(1j * self.pargs[2]) * np.sin(self.pargs[0] / 2)],
            [np.exp(1j * self.pargs[1]) * np.sin(self.pargs[0] / 2),
             np.exp(1j * (self.pargs[1] + self.pargs[2])) * np.cos(self.pargs[0] / 2)]
        ], dtype=self._precision)

    def inverse(self):
        _CU3 = self.copy()
        _CU3.pargs = [self.pargs[0], np.pi - self.pargs[2], np.pi - self.pargs[1]]

        return _CU3

    def build_gate(self):
        from QuICT.qcda.synthesis import UnitaryDecomposition

        assert self.controls + self.targets > 0
        mapping_args = self.cargs + self.targs
        cgate, _ = UnitaryDecomposition().execute(self.matrix)
        if len(mapping_args) == self.controls + self.targets:
            cgate & mapping_args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


CU3 = CU3Gate()


class FSimGate(BasicGate):
    """ fSim gate """
    def __init__(self, params: list = [np.pi / 2, 0]):
        super().__init__(
            controls=0,
            targets=2,
            params=2,
            type_=GateType.fsim,
            matrix_type=MatrixType.ctrl_normal
        )

        self.pargs = params

    def __call__(self, alpha, beta):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate
            beta (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        params = [alpha, beta]
        for param in params:
            self.permit_element(param)

        return FSimGate(params)

    @property
    def matrix(self):
        costh = np.cos(self.pargs[0])
        sinth = np.sin(self.pargs[0])
        phi = self.pargs[1]

        return np.array([
            [1, 0, 0, 0],
            [0, costh, -1j * sinth, 0],
            [0, -1j * sinth, costh, 0],
            [0, 0, 0, np.exp(-1j * phi)]
        ], dtype=self._precision)

    def inverse(self):
        _FSim = self.copy()
        _FSim.pargs = [-self.pargs[0], -self.pargs[1]]

        return _FSim


FSim = FSimGate()


class RxxGate(BasicGate):
    """ Rxx gate """
    def __init__(self, params: list = [0]):
        super().__init__(
            controls=0,
            targets=2,
            params=1,
            type_=GateType.rxx,
            matrix_type=MatrixType.normal_normal
        )

        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return RxxGate([alpha])

    @property
    def matrix(self):
        costh = np.cos(self.parg / 2)
        sinth = np.sin(self.parg / 2)

        return np.array([
            [costh, 0, 0, -1j * sinth],
            [0, costh, -1j * sinth, 0],
            [0, -1j * sinth, costh, 0],
            [-1j * sinth, 0, 0, costh]
        ], dtype=self._precision)

    def inverse(self):
        _Rxx = self.copy()
        _Rxx.pargs = [-self.parg]

        return _Rxx

    def build_gate(self):
        from QuICT.core.gate import CompositeGate

        cgate = CompositeGate()
        with cgate:
            H & 0
            H & 1
            CX & [0, 1]
            Rz(self.parg) & 1
            CX & [0, 1]
            H & 0
            H & 1

        args = self.cargs + self.targs
        if len(args) == self.controls + self.targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


Rxx = RxxGate()


class RyyGate(BasicGate):
    """ Ryy gate """
    def __init__(self, params: list = [np.pi / 2]):
        super().__init__(
            controls=0,
            targets=2,
            params=1,
            type_=GateType.ryy,
            matrix_type=MatrixType.normal_normal
        )

        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return RyyGate([alpha])

    @property
    def matrix(self):
        costh = np.cos(self.parg / 2)
        sinth = np.sin(self.parg / 2)

        return np.array([
            [costh, 0, 0, 1j * sinth],
            [0, costh, -1j * sinth, 0],
            [0, -1j * sinth, costh, 0],
            [1j * sinth, 0, 0, costh]
        ], dtype=self._precision)

    def inverse(self):
        _Ryy = self.copy()
        _Ryy.pargs = [-self.parg]

        return _Ryy

    def build_gate(self):
        from QuICT.core.gate import CompositeGate

        cgate = CompositeGate()
        with cgate:
            Hy & 0
            Hy & 1
            CX & [0, 1]
            Rz(self.parg) & 1
            CX & [0, 1]
            Hy & 0
            Hy & 1

        args = self.cargs + self.targs
        if len(args) == self.controls + self.targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


Ryy = RyyGate()


class RzzGate(BasicGate):
    """ Rzz gate """
    def __init__(self, params: list = [np.pi / 2]):
        super().__init__(
            controls=0,
            targets=2,
            params=1,
            type_=GateType.rzz,
            matrix_type=MatrixType.diag_diag
        )

        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return RzzGate([alpha])

    @property
    def matrix(self):
        expth = np.exp(0.5j * self.parg)
        sexpth = np.exp(-0.5j * self.parg)

        return np.array([
            [sexpth, 0, 0, 0],
            [0, expth, 0, 0],
            [0, 0, expth, 0],
            [0, 0, 0, sexpth]
        ], dtype=self._precision)

    def inverse(self):
        _Rzz = self.copy()
        _Rzz.pargs = [-self.parg]

        return _Rzz

    def build_gate(self):
        from QuICT.core.gate import CompositeGate

        cgate = CompositeGate()
        with cgate:
            CX & [0, 1]
            Rz(self.parg) & 1
            CX & [0, 1]

        args = self.cargs + self.targs
        if len(args) == self.controls + self.targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


Rzz = RzzGate()


class RzxGate(BasicGate):
    """ Rzx gate """
    def __init__(self, params: list = [np.pi / 2]):
        super().__init__(
            controls=0,
            targets=2,
            params=1,
            type_=GateType.rzx,
            matrix_type=MatrixType.diag_normal
        )

        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return RzxGate([alpha])

    @property
    def matrix(self):
        costh = np.cos(self.parg / 2)
        sinth = np.sin(self.parg / 2)

        return np.array([
            [costh, -1j * sinth, 0, 0],
            [-1j * sinth, costh, 0, 0],
            [0, 0, costh, 1j * sinth],
            [0, 0, 1j * sinth, costh]
        ], dtype=self._precision)

    def inverse(self):
        _Rzx = self.copy()
        _Rzx.pargs = [-self.parg]

        return _Rzx

    def build_gate(self):
        from QuICT.core.gate import CompositeGate

        cgate = CompositeGate()
        with cgate:
            H & 0
            CX & [0, 1]
            Rz(self.parg) & 1
            CX & [0, 1]
            H & 0

        args = self.cargs + self.targs
        if len(args) == self.controls + self.targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


Rzx = RzxGate()


class MeasureGate(BasicGate):
    """ z-axis Measure gate

    Measure one qubit along z-axis.
    After acting on the qubit(circuit flush), the qubit get the value 0 or 1
    and the amplitude changed by the result.
    """

    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.measure,
            matrix_type=MatrixType.special
        )

    @property
    def matrix(self) -> np.ndarray:
        raise GateMatrixError("try to get the matrix of measure gate")


Measure = MeasureGate()


class ResetGate(BasicGate):
    """ Reset gate

    Reset the qubit into 0 state,
    which change the amplitude
    """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.reset,
            matrix_type=MatrixType.special
        )

    @property
    def matrix(self) -> np.ndarray:
        raise GateMatrixError("try to get the matrix of reset gate")


Reset = ResetGate()


class BarrierGate(BasicGate):
    """ Barrier gate

    In IBMQ, barrier gate forbid the optimization cross the gate,
    It is invalid in out circuit now.
    """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=1,
            params=0,
            type_=GateType.barrier,
            matrix_type=MatrixType.special
        )

    @property
    def matrix(self) -> np.ndarray:
        raise GateMatrixError("try to get the matrix of barrier gate")


Barrier = BarrierGate()


class SwapGate(BasicGate):
    """ Swap gate

    In the computation, it will not change the amplitude.
    Instead, it change the index of a Tangle.
    """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=2,
            params=0,
            type_=GateType.swap,
            matrix_type=MatrixType.swap
        )

        self.matrix = np.array([
            [1, 0, 0, 0],
            [0, 0, 1, 0],
            [0, 1, 0, 0],
            [0, 0, 0, 1]
        ], dtype=np.complex128)

    def build_gate(self):
        from QuICT.core.gate import CompositeGate

        cgate = CompositeGate()
        with cgate:
            CX & [0, 1]
            CX & [1, 0]
            CX & [0, 1]

        args = self.cargs + self.targs
        if len(args) == self.controls + self.targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


Swap = SwapGate()


class iSwapGate(BasicGate):
    """ iSwap gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=2,
            params=0,
            type_=GateType.iswap,
            matrix_type=MatrixType.swap,
        )

        self.matrix = np.array([
            [1, 0, 0, 0],
            [0, 0, 1j, 0],
            [0, 1j, 0, 0],
            [0, 0, 0, 1]
        ], dtype=np.complex128)


iSwap = iSwapGate()


class iSwapDaggerGate(BasicGate):
    """ iSwap gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=2,
            params=0,
            type_=GateType.iswapdg,
            matrix_type=MatrixType.swap,
        )

        self.matrix = np.array([
            [1, 0, 0, 0],
            [0, 0, -1j, 0],
            [0, -1j, 0, 0],
            [0, 0, 0, 1]
        ], dtype=np.complex128)


iSwap_dagger = iSwapDaggerGate()


class SquareRootiSwapGate(BasicGate):
    """ Square Root of iSwap gate """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=2,
            params=0,
            type_=GateType.sqiswap,
            matrix_type=MatrixType.swap,
        )

        self.matrix = np.array([
            [1, 0, 0, 0],
            [0, 0, (1 + 1j) / np.sqrt(2), 0],
            [0, (1 + 1j) / np.sqrt(2), 0, 0],
            [0, 0, 0, 1]
        ], dtype=np.complex128)


sqiSwap = SquareRootiSwapGate()


# PermGate class -- no qasm
class PermGate(BasicGate):
    """ Permutation gate

    A special gate defined in our circuit,
    It can change an n-qubit qureg's amplitude by permutaion,
    the parameter is a 2^n list describes the permutation.
    """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=0,
            params=0,
            type_=GateType.perm,
            matrix_type=MatrixType.special
        )

    def __call__(self, targets: int, params: list):
        """ pass permutation to the gate

        the length of params must be n, and should be a permutation for [0, n) without repeat

        Args:
            targets(int): the number of qubits
            params(list): the permutation parameters

        Returns:
            PermGate: the gate after filled by parameters
        """
        if not isinstance(params, list):
            raise TypeError("PermGate.params", "list", type(params))
        if not isinstance(targets, int):
            raise TypeError("PermGate.targets", "int", type(targets))

        assert len(params) == targets, GateParametersAssignedError("the length of params must equal to targets")

        _gate = self.copy()
        _gate.targets = targets
        _gate.params = targets
        for idx in params:
            if not isinstance(idx, int):
                raise TypeError("PermGate.params.values", "int", type(idx))
            if idx < 0 or idx >= _gate.targets:
                raise ValueError("PermGate.params.values", f"[0, {targets}]", idx)
            if idx in _gate.pargs:
                raise ValueError("PermGate.params.values", "have no duplicated value", idx)

            _gate.pargs.append(idx)

        return _gate

    def inverse(self):
        _gate = self.copy()
        _gate.pargs = [self.targets - 1 - p for p in self.pargs]

        return _gate

    def build_gate(self, targs: list = None):
        from QuICT.core.gate import CompositeGate

        swap_args: list[list[int]] = perm_decomposition(self.pargs[:])
        cgate = CompositeGate()
        with cgate:
            for swap_arg in swap_args:
                Swap & swap_arg

        if targs is not None:
            assert len(targs) == self.targets + self.controls, \
                GateQubitAssignedError("The qubits number should equal to the quantum gate.")
            cgate & targs

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


Perm = PermGate()


class PermFxGate(BasicGate):
    """ act an Fx oracle on a qureg

    This Class is the subClass of PermGate.
    In fact, we calculate the permutation by the parameters.

    """

    def __init__(self):
        super().__init__(
            controls=0,
            targets=0,
            params=0,
            type_=GateType.perm_fx,
            matrix_type=MatrixType.normal
        )

    def __call__(self, n: int, params: list):
        """ pass Fx to the gate

        Args:
            n (int): the number of targets
            params (list[int]): the list of index, and the index represent which should be 1.

        Returns:
            PermFxGate: the gate after filled by parameters
        """
        if not isinstance(params, list) or not isinstance(n, int):
            raise TypeError(f"n must be int {type(n)}, params must be list {type(params)}")

        N = 1 << n
        for p in params:
            if p >= N:
                raise Exception("the params should be less than N")

        _gate = self.copy()
        _gate.params = 1 << (n + 1)
        _gate.targets = n + 1
        for idx in range(1 << _gate.targets):
            if idx >> 1 in params:
                _gate.pargs.append(idx ^ 1)
            else:
                _gate.pargs.append(idx)

        _gate.matrix = self._build_matrix(_gate.targets, _gate.pargs)

        return _gate

    def _build_matrix(self, targets, pargs):
        matrix_ = np.zeros((1 << targets, 1 << targets), dtype=self.precision)
        for idx, p in enumerate(pargs):
            matrix_[idx, p] = 1

        return matrix_


PermFx = PermFxGate()


class UnitaryGate(BasicGate):
    """ Custom gate

    act an unitary matrix on the qureg,
    the parameters is the matrix

    """
    def __init__(self):
        super().__init__(
            controls=0,
            targets=0,
            params=0,
            type_=GateType.unitary
        )

    def __call__(self, params: np.array, matrix_type: MatrixType = MatrixType.normal):
        """ pass the unitary matrix

        Args:
            params(np.array/list): contain 2^n * 2^n elements, which
            form an unitary matrix.

        Returns:
            UnitaryGateGate: the gate after filled by parameters
        """
        _u = UnitaryGate()

        if isinstance(params, list):
            params = np.array(params, dtype=self._precision)

        matrix_size = params.size
        if matrix_size == 0:
            raise GateMatrixError("the list or tuple passed in shouldn't be empty")

        length, width = params.shape
        if length != width:
            N = int(np.log2(matrix_size))
            assert N ^ 2 == matrix_size, GateMatrixError("the shape of unitary matrix should be square.")

            params = params.reshape(N, N)

        n = int(np.log2(params.shape[0]))
        if (1 << n) != params.shape[0]:
            raise GateMatrixError("the length of list should be the square of power(2, n)")

        _u.targets = n
        _u.matrix = params.astype(self._precision)
        if n <= 3:
            _u._validate_matrix_type()
        else:
            _u._matrix_type = matrix_type

        return _u

    def _validate_matrix_type(self):
        if self._is_diagonal():
            is_control = np.allclose(self.matrix[:-1, :-1], np.identity((2 ** self.targets - 1), dtype=self._precision))
            self._matrix_type = MatrixType.control if is_control else MatrixType.diagonal

        if (
            np.allclose(self.matrix[:-2, :-2], np.identity((2 ** self.targets - 2), dtype=self._precision)) and
            np.sum(self.matrix[:-2, -1]) + self.matrix[-1, -1] == 0 and
            np.sum(self.matrix[-1, :-2]) + self.matrix[-2, -2] == 0
        ):
            self._matrix_type = MatrixType.reverse

    def copy(self):
        gate = super().copy()
        gate.matrix = self.matrix

        return gate

    def inverse(self):
        _U = super().copy()
        inverse_matrix = np.array(
            np.mat(self._matrix).reshape(1 << self.targets, 1 << self.targets).H.reshape(1, -1),
            dtype=self._precision
        )
        _U.matrix = inverse_matrix
        _U.targets = self.targets

        return _U

    def build_gate(self):
        from QuICT.qcda.synthesis import UnitaryDecomposition

        assert self.controls + self.targets > 0
        mapping_args = self.cargs + self.targs
        cgate, _ = UnitaryDecomposition().execute(self.matrix)
        cgate & mapping_args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


Unitary = UnitaryGate()


class CCXGate(BasicGate):
    """ Toffoli gate

    When using this gate, it will be showed as a whole gate
    instend of being split into smaller gate

    """
    def __init__(self):
        super().__init__(
            controls=2,
            targets=1,
            params=0,
            type_=GateType.ccx,
            matrix_type=MatrixType.reverse
        )

        self.matrix = np.array([
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0, 1, 0]
        ], dtype=self._precision)

        self._target_matrix = np.array([
            [0, 1],
            [1, 0]
        ], dtype=self._precision)

    @property
    def target_matrix(self) -> np.ndarray:
        return self._target_matrix

    def build_gate(self):
        from QuICT.core.gate import CompositeGate

        cgate = CompositeGate()
        with cgate:
            H & 2
            CX & [2, 1]
            T_dagger & 1
            CX & [0, 1]
            T & 1
            CX & [2, 1]
            T_dagger & 1
            CX & [0, 1]
            T & 1
            CX & [0, 2]
            T_dagger & 2
            CX & [0, 2]
            T & 0
            T & 2
            H & 2

        args = self.cargs + self.targs
        if len(args) == self.controls + self.targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


CCX = CCXGate()


class CCZGate(BasicGate):
    """ Multi-control Z gate

    When using this gate, it will be showed as a whole gate
    instend of being split into smaller gate

    """
    def __init__(self):
        super().__init__(
            controls=2,
            targets=1,
            params=0,
            type_=GateType.ccz
        )

        self.matrix = np.array([
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, -1]
        ], dtype=self._precision)

        self._target_matrix = np.array([
            [1, 0],
            [0, -1]
        ], dtype=self._precision)

    @property
    def target_matrix(self) -> np.ndarray:
        return self._target_matrix

    def build_gate(self):
        from QuICT.core.gate import CompositeGate

        cgate = CompositeGate()
        with cgate:
            CX & [2, 1]
            T_dagger & 1
            CX & [0, 1]
            T & 1
            CX & [2, 1]
            T_dagger & 1
            CX & [0, 1]
            T & 1
            CX & [0, 2]
            T_dagger & 2
            CX & [0, 2]
            T & 0
            T & 2

        args = self.cargs + self.targs
        if len(args) == self.controls + self.targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


CCZ = CCZGate()


class CCRzGate(BasicGate):
    """ controlled-Rz gate with two control bits """
    def __init__(self, params: list = [0]):
        super().__init__(
            controls=2,
            targets=1,
            params=1,
            type_=GateType.ccrz,
            matrix_type=MatrixType.diagonal
        )

        self.pargs = params

    def __call__(self, alpha):
        """ Set parameters for the gate.

        Args:
            alpha (int/float/complex): The parameter for gate

        Raises:
            TypeError: param not one of int/float/complex

        Returns:
            BasicGate: The gate with parameters
        """
        self.permit_element(alpha)

        return CCRzGate([alpha])

    @property
    def matrix(self):
        return np.array([
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, np.exp(-self.parg / 2 * 1j), 0],
            [0, 0, 0, 0, 0, 0, 0, np.exp(self.parg / 2 * 1j)]
        ], dtype=self._precision)

    @property
    def target_matrix(self) -> np.ndarray:
        return np.array([
            [np.exp(-self.parg / 2 * 1j), 0],
            [0, np.exp(self.parg / 2 * 1j)]
        ], dtype=self._precision)

    def inverse(self):
        _CCRz = self.copy()
        _CCRz.pargs = -self.parg
        return _CCRz

    def build_gate(self):
        from QuICT.core.gate import CompositeGate

        cgate = CompositeGate()
        with cgate:
            CRz(self.parg / 2) & [1, 2]
            CX & [0, 1]
            CRz(-self.parg / 2) & [1, 2]
            CX & [0, 1]
            CRz(self.parg / 2) & [0, 2]

        args = self.cargs + self.targs
        if len(args) == self.controls + self.targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


CCRz = CCRzGate()


class QFTGate(BasicGate):
    """ QFT gate """
    @property
    def matrix(self) -> np.ndarray:
        if self._matrix is None:
            cgate = self.build_gate()
            self._matrix = cgate.matrix()
        return self._matrix

    def __init__(self, targets: int = 3):
        super().__init__(
            controls=0,
            targets=targets,
            params=0,
            type_=GateType.qft
        )

    def __call__(self, targets: int):
        """ pass the unitary matrix

        Args:
            targets(int): point out the number of bits of the gate

        Returns:
            QFTGate: the QFTGate after filled by target number
        """
        return QFTGate(targets)

    def inverse(self):
        _IQFT = IQFTGate()
        _IQFT.targs = copy.deepcopy(self.targs)
        _IQFT.targets = self.targets
        return _IQFT

    def build_gate(self, targets: int = 0):
        from QuICT.core.gate import CompositeGate

        if targets == 0:
            targets = self.targets

        cgate = CompositeGate()
        with cgate:
            for i in range(targets):
                H & i
                for j in range(i + 1, targets):
                    CU1(2 * np.pi / (1 << j - i + 1)) & [j, i]

        args = self.cargs + self.targs
        if len(args) == targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


QFT = QFTGate()


class IQFTGate(QFTGate):
    """ IQFT gate """
    def __call__(self, targets: int):
        """ pass the unitary matrix

        Args:
            targets(int): point out the number of bits of the gate

        Returns:
            IQFTGate: the IQFTGate after filled by target number
        """
        return IQFTGate(targets)

    def inverse(self):
        _QFT = QFTGate()
        _QFT.targs = copy.deepcopy(self.targs)
        _QFT.targets = self.targets
        return _QFT

    def build_gate(self, targets: int = 0):
        from QuICT.core.gate import CompositeGate

        if targets == 0:
            targets = self.targets

        cgate = CompositeGate()
        with cgate:
            for i in range(targets - 1, -1, -1):
                for j in range(targets - 1, i, -1):
                    CU1(-2 * np.pi / (1 << j - i + 1)) & [j, i]
                H & i

        args = self.cargs + self.targs
        if len(args) == targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


IQFT = IQFTGate()


class CSwapGate(BasicGate):
    """ Fredkin gate

    When using this gate, it will be showed as a whole gate
    instend of being split into smaller gate
    """
    def __init__(self):
        super().__init__(
            controls=1,
            targets=2,
            params=0,
            type_=GateType.cswap,
            matrix_type=MatrixType.swap
        )

        self.matrix = np.array([
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1]
        ], dtype=self._precision)

        self._target_matrix = np.array([
            [1, 0, 0, 0],
            [0, 0, 1, 0],
            [0, 1, 0, 0],
            [0, 0, 0, 1]
        ], dtype=self._precision)

    @property
    def target_matrix(self) -> np.ndarray:
        return self._target_matrix

    def build_gate(self):
        from QuICT.core.gate import CompositeGate

        cgate = CompositeGate()
        with cgate:
            CX & [2, 1]
            H & 2
            CX & [2, 1]
            T_dagger & 1
            CX & [0, 1]
            T & 1
            CX & [2, 1]
            T_dagger & 1
            CX & [0, 1]
            T & 1
            CX & [0, 2]
            T_dagger & 2
            CX & [0, 2]
            T & 0
            T & 2
            H & 2
            CX & [2, 1]

        args = self.cargs + self.targs
        if len(args) == self.controls + self.targets:
            cgate & args

        if self._precision == np.complex64:
            cgate.convert_precision()

        return cgate


CSwap = CSwapGate()



import os
import datetime
import json
import jwt
import requests

from .encrypt_manager import EncryptManager


class EncryptedRequest:
    """ The class contains encrypted request(get, post, delete). """
    def __init__(self):
        self._encrypt = EncryptManager()
        self.__SALT = "TestForQuICT"

    def get(self, url: str, user_info: tuple):
        """ Send a GET Request.

        Args:
            url (str): URL for request.
            user_info (tuple): the login user's information.
        """
        aes_key = os.urandom(16)
        header = self._generate_header(aes_key, user_info)

        # Get response.
        response = requests.get(
            url=url,
            headers=header
        )

        # decrepted response
        return self._decrepted_response(response, user_info)

    def post(
        self,
        url: str,
        json_dict: dict = None,
        user_info: tuple = None,
        is_login: bool = False
    ):
        """ Send a POST Request.

        Args:
            url (str): URL for request.
            json_dict (dict, optional): The send data. Defaults to None.
            user_info (tuple, optional): The login user's information. Defaults to None.
            is_login (bool, optional): whether is login post or not. Defaults to False.
        """
        aes_key = os.urandom(16)
        header = self._generate_header(aes_key, user_info, is_login)
        if json_dict is not None:
            json_dict = json.dumps(json_dict)

        response = requests.post(
            url=url,
            headers=header,
            data=self._encrypt.encryptedmsg(json_dict, aes_key) if json_dict is not None else None
        )

        # decrepted response
        return self._decrepted_response(response, user_info, is_login)

    def delete(self, url: str, json_dict: dict = None, user_info: tuple = None):
        """ Send a DELETE Request.

        Args:
            url (str): URL for request.
            json_dict (dict, optional): The send data. Defaults to None.
            user_info (tuple, optional): The login user's information. Defaults to None.
        """
        aes_key = os.urandom(16)
        header = self._generate_header(aes_key, user_info)
        if json_dict is not None:
            json_dict = json.dumps(json_dict)

        response = requests.delete(
            url=url,
            headers=header,
            data=self._encrypt.encryptedmsg(json_dict, aes_key) if json_dict is not None else None
        )

        # decrepted response
        return self._decrepted_response(response, user_info)

    def _generate_header(self, aes_key: bytes, user_info: tuple, is_login: bool = False):
        """ Generate Header for Request. Using AES and HRS to encrypte.

        Args:
            aes_key (bytes): The encrypted key for AES Mode.
            user_info (tuple): The login user's information
            is_login (bool, optional): whether is login post or not. Defaults to False.

        Returns:
            dict: The JWT token.
        """
        username = user_info[1]
        password = user_info[2][:16] if not is_login else self.__SALT

        payload = {
            'username': username,
            'aes_key': self._encrypt.encryptedmsg(aes_key, password).decode('ascii'),
            'exp': datetime.datetime.utcnow() + datetime.timedelta(minutes=15)
        }

        jwt_token = jwt.encode(
            payload=payload,
            key=self.__SALT,
            algorithm="HS256"
        )

        return {"Authorization": f"Bearer {jwt_token}"}

    def _decrepted_response(self, response: requests.Response, user_info: tuple, is_login: bool = False):
        jwt_token = response.headers.get('Authorization')
        if jwt_token and jwt_token.startswith('Bearer '):
            payload = jwt.decode(jwt_token[7:], self.__SALT, ["HS256"])
        else:
            payload = None

        username = user_info[1]
        password = user_info[2][:16] if not is_login else self.__SALT
        assert username == payload.get("username", None)
        encrypted_aes_key = payload.get('aes_key')

        content = response.content
        if not content:
            return None

        aes_key = self._encrypt.decryptedmsg(encrypted_aes_key, password, True)
        decrypted_data = json.loads(self._encrypt.decryptedmsg(content, aes_key))

        if isinstance(decrypted_data, dict) and 'error' in decrypted_data.keys():
            raise KeyError(f"{decrypted_data['error']}")

        return decrypted_data



# Author  : Zhu Qinlin

"""
The (2n+2)-qubit circuit used in the Shor algorithm is designed \
by THOMAS HANER, MARTIN ROETTELER, and KRYSTA M. SVORE in \
"Factoring using 2n+2 qubits with Toffoli based modular multiplication"
"""

from fractions import Fraction
from math import pi
import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.qcda.synthesis.arithmetic.hrs import *
from QuICT.simulation.state_vector import CircuitSimulator
from QuICT.core.operator import Trigger, CheckPoint
from .utility import *

from QuICT.tools import Logger
from QuICT.tools.exception.core import *

logger = Logger("HRS-zip")


def construct_circuit(a: int, N: int, eps: float = 1 / 10):
    # phase estimation procedure
    n = int(np.ceil(np.log2(N + 1)))
    t = int(2 * n + 1 + np.ceil(np.log(2 + 1 / (2 * eps))))
    logger.info(f"\tcircuit construction begin: circuit: n = {n} t = {t}")

    circuit = Circuit(2 * n + 2)
    x_reg = list(range(n))
    trickbit = [2 * n + 1]

    # cgates[measured qubit index][added gate index][measured result]
    cgates = [
        [[CompositeGate() for measured in range(2)] for j in range(t + 1)]
        for i in range(t)
    ]
    # triggers[measured qubit index][added gate index]
    triggers = [
        [
            Trigger(1, [cgates[i][j][measured] for measured in range(2)])
            for j in range(t + 1)
        ]
        for i in range(t)
    ]
    triggers_reset = list(range(t))
    # checkpoints[measured qubit index]
    checkpoints = [CheckPoint() for i in range(t)]

    # triggers!!!
    for k in range(t):  # end at k-th checkpoints
        for i in range(k):  # start at i-th qubits (compressed)
            cpc = checkpoints[k].get_child()
            Rz(-np.pi / (1 << (k - i))) | cgates[i][k - 1 - i][1](trickbit)
            cpc | cgates[i][k - 1 - i][1]
    # triggers for reset!!!
    for i in range(t):
        cgate = CompositeGate()
        X | cgate(trickbit)
        triggers_reset[i] = Trigger(1, [CompositeGate(), cgate])

    # subcircuit: init \ket{1}\ket{0}
    X | circuit(x_reg[n - 1])

    history_indices = []
    ptr = 0
    for k in range(t):
        # subcircuit CUa
        H | circuit(trickbit)
        gate_pow = pow(a, 1 << (t - 1 - k), N)
        CHRSMulMod.execute(n, gate_pow, N) | circuit
        # subcircuit: semi-classical QFT
        checkpoints[k] | circuit
        H | circuit(trickbit)
        # subcircuit: measure & reset trickbit
        for i in range(t - k):
            if i == 0:
                history_indices.append(ptr)
            triggers[k][i] | circuit(trickbit)
            ptr += 1
        triggers_reset[k] | circuit(trickbit)
        ptr += 1
    return circuit, trickbit + history_indices[::-1]


def order_finding(a: int, N: int, eps: float = 1 / 10, simulator=CircuitSimulator()):
    """
    Shor algorithm by THOMAS HANER, MARTIN ROETTELER, and KRYSTA M. SVORE \
    in "Factoring using 2n+2 qubits with Toffoli based modular multiplication"
    Quantum algorithm to compute the order of a (mod N), when gcd(a,N)=1.
    """
    # phase estimation procedure
    n = int(np.ceil(np.log2(N + 1)))
    t = int(2 * n + 1 + np.ceil(np.log(2 + 1 / (2 * eps))))
    logger.info(f"\torder_finding begin: circuit: n = {n} t = {t}")
    trickbit_store = [0] * t

    circuit = Circuit(2 * n + 2)
    x_reg = list(range(n))
    ancilla = list(range(n, 2 * n))
    indicator = 2 * n
    trickbit = [2 * n + 1]
    # subcircuit: init \ket{1}\ket{0}
    circuit = Circuit(2 * n + 2)
    X | circuit(x_reg[n - 1])
    simulator.run(circuit)

    for k in range(t):
        # subcircuit CUa
        circuit = Circuit(2 * n + 2)
        H | circuit(trickbit)
        gate_pow = pow(a, 1 << (t - 1 - k), N)
        CHRSMulMod.execute(n, gate_pow, N) | circuit
        simulator.run(circuit, use_previous=True)
        # subcircuit: semi-classical QFT
        circuit = Circuit(2 * n + 2)
        for i in range(k):
            if trickbit_store[i]:
                Rz(-pi / (1 << (k - i))) | circuit(trickbit)
        H | circuit(trickbit)
        simulator.run(circuit, use_previous=True)
        circuit = Circuit(2 * n + 2)
        for idx in [trickbit] + ancilla + [indicator]:
            Measure | circuit(idx)
        simulator.run(circuit, use_previous=True)
        # subcircuit: measure & reset trickbit
        assert int(circuit[indicator]) == 0 and int(circuit[ancilla]) == 0
        logger.info(f"\tthe {k}th trickbit measured to be {int(circuit[trickbit])}")
        trickbit_store[k] = int(circuit[trickbit])
        if trickbit_store[k] == 1:
            circuit = Circuit(2 * n + 2)
            X | circuit(trickbit)
            simulator.run(circuit, use_previous=True)

    # for idx in x_reg: Measure | circuit(idx)
    trickbit_store.reverse()
    logger.info(f"\tphi~ (approximately s/r) in binary form is {trickbit_store}")
    # continued fraction procedure
    phi_ = sum([(trickbit_store[i] * 1.0 / (1 << (i + 1))) for i in range(t)])
    logger.info(f"\tphi~ (approximately s/r) in decimal form is {phi_}")
    if phi_ == 0.0:
        logger.info("\torder_finding failed: phi~ = 0")
        return 0
    frac = Fraction(phi_).limit_denominator(N)
    logger.info(f"\tContinued fraction expansion of phi~ is {frac}")
    return frac.denominator



from enum import Enum
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor, as_completed

import cupy as cp
import numpy as np
from cupy.cuda import nccl

from QuICT.utility import Proxy
from QuICT.core import Circuit
from .transpile import Transpile
from .multi_nodes_simulator import MultiNodesSimulator
from QuICT.ops.linalg.gpu_calculator import VectorPermutation
from QuICT.simulation.utils import options_validation


if mp.get_start_method(allow_none=True) != "spawn":
    mp.set_start_method("spawn", force=True)


class DeviceType(Enum):
    cpu = "CPU",
    gpu = "GPU"


class ModeType(Enum):
    local = "local"
    distributed = "distributed"


def worker(circuit, ndev, uid, dev_id, options, device: str = "GPU"):
    proxy = Proxy(ndevs=ndev, uid=uid, dev_id=dev_id)
    simulator = MultiNodesSimulator(
        proxy=proxy,
        device=device,
        gpu_id=dev_id,
        **options
    )
    state = simulator.run(circuit)

    return dev_id, state


class MultiNodesController:
    # Not support noise circuit
    def __init__(
        self,
        ndev: int,
        dev_type: DeviceType = DeviceType.gpu,
        mode: ModeType = ModeType.local,
        **options
    ):
        self.ndev = ndev
        self._device_type = dev_type
        self._mode_type = mode
        if options_validation(options=options, device="GPU", backend="state_vector"):
            self._options = options
        else:
            raise KeyError("Unmatched options arguments depending on GPU and state_vector.")

        self._transpiler = Transpile(self.ndev)

    def run(self, circuit: Circuit):
        # transpile circuit
        divided_circuits, split_qubit = self._transpiler.run(circuit)

        # Prepare mapping permutation
        inner_mapping = list(range(circuit.width()))
        inner_mapping.remove(int(split_qubit))
        permutation = [split_qubit] + inner_mapping

        # start launch multi-node simulator
        if self._mode_type == ModeType.local:
            return self._launch_local(divided_circuits, permutation)

    def _launch_local(self, dcircuit: Circuit, permutation: list):
        # Using multiprocess to start simulators, only for GPUs
        proxy_id = nccl.get_unique_id()
        with ProcessPoolExecutor(max_workers=self.ndev) as executor:
            tasks = [
                executor.submit(
                    worker,
                    dcircuit,
                    self.ndev,
                    proxy_id,
                    dev_id,
                    self._options
                ) for dev_id in range(self.ndev)
            ]

        results = []
        for t in as_completed(tasks):
            results.append(t.result())

        z = [None] * self.ndev
        for idx, vec in results:
            z[idx] = vec

        combined_sv = cp.concatenate(z)
        return VectorPermutation(combined_sv, np.array(permutation))

    def _launch_distributed(self):
        # send job to distributed
        pass



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/11/5 8:55
# @Author  : Han Yu
# @File    : cnot_ancillae.py

from math import log2, ceil, floor, sqrt
from typing import Union

import numpy as np

from QuICT.core import *
from QuICT.core.gate import CX, CompositeGate, GateType
from QuICT.qcda.utility import OutputAligner
from QuICT.tools.exception.core import ValueError, CircuitAppendError


class CnotAncilla(object):
    """ Optimization the circuit by (3s+1)n ancillary qubits

    Optimal Space-Depth Trade-Off of CNOT Circuits in Quantum Logic Synthesis
    Theorem 7
    parallelize any n-qubit CNOT circuit into O(n/slogn) depth with (3s+1) ancillas,
    where 1 <= s <= O(n/log^2n). time complex is \tilde{O}(n^w), which is time for get inverse of a matrix.
    https://arxiv.org/pdf/1907.05087.pdf
    """
    def __init__(self, size=1):
        """
        Args:
            size(int): the 's' in the Theorem, where (3s+1) ancillas would be used
        """
        self.size = size

    @OutputAligner()
    def execute(self, circuit: Union[Circuit, CompositeGate]):
        """
        Args:
            circuit(CompositeGate/Circuit): circuit to be optimize

        Returns:
            CompositeGate/Circuit: the optimized circuit
        """
        # transform the CNOT circuit into a matrix with 0 and 1
        self.width = circuit.width()
        if self.width < 4:
            raise ValueError("CnotAncilla.circuit.width", ">= 4", self.width)
        matrix = np.identity(self.width, dtype=bool)
        for gate in circuit.gates:
            if gate.type != GateType.cx:
                raise CircuitAppendError(f"the input circuit should be a CNOT circuit, but it contains {str(gate)}")
            matrix[gate.targs] = np.bitwise_xor(matrix[gate.cargs], matrix[gate.targs])

        # apply Theorem 7 to build new circuit
        self.CNOT = []
        # x, z and c mean three part in the circuit in the theorem 7
        x: list = [j for j in range(self.width)]
        z: list = [j + self.width for j in range(self.width)]
        c: list = [j + 2 * self.width for j in range(3 * self.size * self.width)]
        inv_matrix = self.inverse_matrix_f2(matrix, self.width)
        # do C_1 operate
        self.lemma_4(matrix, x, c, z)
        # do C_2 operate
        self.lemma_4(inv_matrix, z, c, x)

        new_circuit = Circuit(circuit.width() * (2 + 3 * self.size))
        for cnot in self.CNOT:
            CX | new_circuit([cnot[0], cnot[1]])
        return new_circuit

    def lemma_4(self, M, x, c, z):
        """ apply Lemma4 to make (x, z, 0) -> (x, z xor Mx, 0)

        time complex: \tilde{O}(n^2)
        depth : O(n/slogn)

        Args:
            M(np.matrix): a matrix in F_2
            x(list<int>): the indexes of first register, length is n
            c(list<int>): the indexes of ancillary register, length is n
            z(list<int>): the indexes of second register, length is 3 * s * n
        """
        # divide the part into size slog^2n and apply Lemma 5
        # we ensure the log2n be even and 2 (sqrt(n) - 1) log2n <= n
        log2n = floor(log2(self.width))
        while log2n % 2 == 1 or 2 * (round(pow(2, log2n / 2)) - 1) * log2n > self.width:
            log2n -= 1
        t = round(log2n * log2n)
        for i in range(ceil(self.width / (self.size * t))):
            self.generate_y_part(M[:, i * self.size * t:], x[i * t * self.size:], c, i, t, z)

    def generate_y_part(self, Y_part, x, c, index, length, z):
        """ :: apply Corollary 3 to make (x, z, 0) -> (x, (T_part)z, 0)

        time complex: \tilde{O}(sn)
        depth : O(logn)

        Args:
            Y_part(np.matrix): n * n, matrix in F_2. We only use the part
                            Y_part[: , :slog^2n]
            x(list<int>): the indexes of first register, we only use x[:slog^2n]
            c(list<int>): the indexes of ancillary register, length is 3 * s * n
            index(int): the index of the Y part
            length(int): the reality value for log^2n
            z(list<int>): the indexes of second register, length is n
        """
        # operate R_a in Corollary 3
        init_len = len(self.CNOT)
        for i in range(self.size):
            if index * length * self.size + i * length < self.width:
                self.generate_y_base(Y_part[:, i * length:], c[i * 3 * self.width:], length, x[i * length:])
        end_len = len(self.CNOT)
        # operate Add in Corollary 3
        for i in range(self.size):
            for j in range(self.width):
                self.CNOT.append((c[i * 3 * self.width + j], z[j]))
        # operate R_a^{-1} in Corollary 3
        self.inverse_gates(init_len, end_len)

    def generate_y_base(self, Y_part, c, length, x):
        """ Apply Lemma 6 to copy part of Y_part in ancillary register

        We generator Y_part[:, :log^2n]
        to c[:n] with the help of c[n :  3 * n]
        time complex: \tilde{O}(n)
        depth : O(logn)

        Args:
            Y_part(np.matrix): Y_part
            c(list<int>): indexes of ancillary register,
            length: the value of log^2n
            x(list<int>): indexes of first register, length is log^2n

        """
        # ancillary qubits in Lemma 6
        ancillary = c[self.width: 3 * self.width]
        # goal qubits in Lemma 6
        ystart = c[:self.width]
        d2logn = int(floor(round(sqrt(length)) / 2))
        sqrtn = int(pow(2, d2logn))
        # real length
        length = min(length, np.shape(Y_part)[1])

        Step1_start = len(self.CNOT)
        for j in range(ceil(length / d2logn)):
            # Step 1 Construct Pj's (Lemma 5)
            cols = min(d2logn, length - j * d2logn)
            r_sqrtn = int(pow(2, cols))
            self.construct_pj(ancillary[sqrtn + sqrtn * d2logn * j // 2:], x[j * d2logn:],
                              ancillary[j * (sqrtn - 1):], r_sqrtn, cols)
        Step1_end = len(self.CNOT)

        pointer = self.width
        for k in range(ceil(length / d2logn)):
            # Step 2 Copy rows in Pj's
            Step2_start = len(self.CNOT)
            cols = min(d2logn, length - k * d2logn)
            r_sqrtn = int(pow(2, cols))
            sl = [0] * (r_sqrtn - 1)
            sl_origin = [[] * 0 for _ in range(r_sqrtn - 1)]
            for u in range(self.width):
                l = 0
                for i in range(cols):
                    if Y_part[u, d2logn * k + i]:
                        l += 1 << i
                if l == 0:
                    continue
                sl[l - 1] += 1
            for i in range(1, r_sqrtn):
                if sl[i - 1] > 0:
                    total = ceil(sl[i - 1] / (4 * d2logn)) - 1
                    number_a = 1
                    now = k * (sqrtn - 1) + i - 1
                    sl_origin[i - 1].append((4 * d2logn, now))
                    while total > 0:
                        for _ in range(number_a):
                            self.CNOT.append((ancillary[now], ancillary[pointer]))
                            sl_origin[i - 1].append((4 * d2logn, pointer))
                            pointer += 1
                            total -= 1
                            if total == 0:
                                break
                            self.CNOT.append((ancillary[now], ancillary[pointer]))
                            sl_origin[i - 1].append((4 * d2logn, pointer))
                            pointer += 1
                            total -= 1
                            if total == 0:
                                break
                            if now == k * (sqrtn - 1) + i - 1:
                                now = pointer - 2
                            else:
                                now += 1
            Step2_end = len(self.CNOT)
            # Step 3
            for u in range(self.width):
                l = 0
                for i in range(cols):
                    if Y_part[u, d2logn * k + i]:
                        l += 1 << i
                if l == 0:
                    continue
                l -= 1
                assert len(sl_origin[l]) > 0
                if sl_origin[l][0][0] == 0:
                    sl_origin[l] = sl_origin[l][1:]
                assert len(sl_origin[l]) > 0
                sl_origin[l][0] = (sl_origin[l][0][0] - 1, sl_origin[l][0][1])
                self.CNOT.append((ancillary[sl_origin[l][0][1]], ystart[u]))
            # Step 4.1 restore of Step2
            self.inverse_gates(Step2_start, Step2_end)
        # Step 4.2 restore of Step1
        self.inverse_gates(Step1_start, Step1_end)

    def construct_pj(self, c, x, z, sqrtn, d2logn):
        """
        Apply Lemma 5 to make (x, z, 0) -> (x, (Pj)z, 0)$$
        Pj is the sqrt(n) * logn/2 matrix go through F_2^{logn / 2}

        Args:
            c(list<int>): the indexes of ancillae, length is sqrt(n) * log2n / 2
            x(list<int>): the indexes of first register, length is log2n / 2
            z(list<int>): the indexes of second register, length is log2n / 2
            sqrtn(int)  : the value of sqrt(n)
            d2logn(int) : the value of logn / 2
        """
        t = []
        total = 0
        first = 0

        start = len(self.CNOT)
        for j in range(d2logn):
            tj = 0
            for i in range(1, sqrtn):
                if i & (1 << j):
                    tj += 1
            total += tj
            t.append(total - 1)
            if tj > 0:
                tj -= 1
                self.CNOT.append((x[j], c[first]))
                now = first
                first += 1
                number_a = 1
                while tj > 0:
                    for i in range(number_a):
                        self.CNOT.append((c[now + i], c[first]))
                        first += 1
                        tj -= 1
                        if tj == 0:
                            break
                        self.CNOT.append((c[now + i], c[first]))
                        first += 1
                        tj -= 1
                        if tj == 0:
                            break
        end = len(self.CNOT)
        for j in range(d2logn):
            for i in range(1, sqrtn):
                if i & (1 << j):
                    self.CNOT.append((c[t[j]], z[i - 1]))
                    t[j] -= 1
        self.inverse_gates(start, end)

    def inverse_gates(self, start, end):
        """ apply the inverse of gate list in CNOT[start:end] on the circuit

        Args:
            start(int): the start index of interval
            end(int): the end index of interval
        """

        for i in range(end - 1, start - 1, -1):
            self.CNOT.append((self.CNOT[i][0], self.CNOT[i][1]))

    @staticmethod
    def inverse_matrix_f2(mat, n):
        """ get the inverse matrix of mat in F_2

        Args:
            mat(np.ndarray): the matrix in F_2
            n(int): the size of the matrix

        Returns:
            np.ndarray: the inverse of mat in F_2
        """
        b = np.zeros(2 * n * n, dtype=bool).reshape(n, 2 * n)
        for i in range(n):
            for j in range(n):
                b[i][j] = mat[i][j]
                b[i][j + n] = False
            b[i][i + n] = True
        for i in range(n):
            if not b[i][i]:
                flag = False
                for j in range(i + 1, n):
                    if b[j][i]:
                        flag = True
                        for t in range(2 * n):
                            b[j][t], b[i][t] = b[i][t], b[j][t]
                if not flag:
                    return False
            for j in range(i + 1, n):
                if b[j][i]:
                    for t in range(2 * n):
                        b[j][t] = b[j][t] ^ b[i][t]
        for i in range(n):
            for j in range(n):
                b[i][j] = b[i][j + n]
        return b[:, :n]



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2022/1/17 9:41
# @Author  : Han Yu, Kaiqi Li
# @File    : circuit.py
from typing import Union, List
import numpy as np

from QuICT.core.qubit import Qubit, Qureg
from QuICT.core.layout import Layout, SupremacyLayout
from QuICT.core.gate import BasicGate, H, Measure, build_random_gate, build_gate
from QuICT.core.utils import (
    GateType,
    CircuitBased,
    unique_id_generator
)
from QuICT.core.operator import (
    Trigger,
    CheckPoint,
    Operator,
    CheckPointChild,
    NoiseGate
)
from .dag_circuit import DAGCircuit

from QuICT.tools import Logger
from QuICT.tools.exception.core import *


logger = Logger("circuit")


class Circuit(CircuitBased):
    """ Implement a quantum circuit

    Circuit is the core part of the framework.

    Attributes:
        wires(Union[Qureg, int]): the number of qubits for the circuit.
        name(str): the name of the circuit.
        topology(list<tuple<int, int>>):
            The topology of the circuit. When the topology list is empty, it will be seemed as fully connected.
        fidelity(float): the fidelity of the circuit
        ancillae_qubits(list<int>): The indexes of ancilla qubits for current circuit.
    """
    @property
    def qubits(self) -> Qureg:
        return self._qubits

    @property
    def ancillae_qubits(self) -> List[int]:
        return self._ancillae_qubits

    @ancillae_qubits.setter
    def ancillae_qubits(self, ancillae_qubits: List[int]):
        for idx in ancillae_qubits:
            if idx < 0 or idx >= self.width():
                raise IndexExceedError(
                    "circuit.ancillae_qubits", [0, self.width()], idx
                )

            self._ancillae_qubits.append(idx)

    @property
    def topology(self) -> Layout:
        return self._topology

    @topology.setter
    def topology(self, topology: Layout):
        if topology is None:
            self._topology = None
            return

        if not isinstance(topology, Layout):
            raise TypeError(
                "Circuit.topology", "Layout", type(topology)
            )

        if topology.qubit_number != self.width():
            raise ValueError(
                "Circuit.topology.qubit_number", self.width(), topology.qubit_number
            )

        self._topology = topology

    @property
    def fidelity(self) -> float:
        return self._fidelity

    @fidelity.setter
    def fidelity(self, fidelity):
        if fidelity is None:
            self._fidelity = None
            return

        if not isinstance(fidelity, float):
            raise TypeError(
                "Circuit.fidelity", "float", type(fidelity)
            )

        if fidelity < 0 or fidelity > 1.0:
            raise ValueError(
                "Circuit.fidelity", "within [0, 1]", {fidelity}
            )

        self._fidelity = fidelity

    def __init__(
        self,
        wires,
        name: str = None,
        topology: Layout = None,
        fidelity: float = None,
        ancillae_qubits: List[int] = None
    ):
        if name is None:
            name = "circuit_" + unique_id_generator()

        super().__init__(name)
        self._ancillae_qubits = []
        self._topology = None
        self._fidelity = None
        self._checkpoints = []

        if isinstance(wires, Qureg):
            self._qubits = wires
        else:
            self._qubits = Qureg(wires)

        if ancillae_qubits is not None:
            self.ancillae_qubits = ancillae_qubits

        logger.debug(f"Initial Quantum Circuit {name} with {len(self._qubits)} qubits.")
        if topology is not None:
            self.topology = topology
            logger.debug(f"The Layout for Quantum Circuit is {self._topology}.")

        if fidelity is not None:
            self.fidelity = fidelity
            logger.debug(f"The Fidelity for Quantum Circuit is {self.fidelity}.")

    def __del__(self):
        """ release the memory """
        self.gates = None
        self._qubits = None
        self.topology = None
        self.fidelity = None
        logger.debug(f"Delete Quantum Circuit {self._name}.")

    def __or__(self, targets):
        """deal the operator '|'

        Use the syntax "circuit | circuit"
        to add the gate of circuit into the circuit/qureg/qubit

        Note that the order of qubits is that control bits first
        and target bits followed.

        Args:
            targets: the targets the gate acts on, it can have following form,
                1) Circuit
        Raise:
            TypeError: the type of targets is wrong
        """
        if not isinstance(targets, Circuit):
            raise TypeError(
                "Circuit.or", "Circuit", type(targets)
            )

        if not self.qubits == targets.qubits:
            diff_qubits = targets.qubits.diff(self.qubits)
            targets.update_qubit(diff_qubits, is_append=True)

        targets.extend(self.gates)

    ####################################################################
    ############         Circuit Qubits Operators           ############
    ####################################################################
    def __call__(self, indexes: object):
        """ get a smaller qureg from this circuit

        Args:
            indexes: the indexes passed in, it can have follow form:
                1) int
                2) list<int>
                3) Qubit
                4) Qureg
        Returns:
            Qureg: the qureg correspond to the indexes
        Exceptions:
            TypeError: the type of indexes is error.
        """
        if isinstance(indexes, int) or type(indexes) is list:
            indexes = self.qubits(indexes)

        if isinstance(indexes, Qubit):
            indexes = Qureg(indexes)

        if not isinstance(indexes, Qureg):
            raise TypeError(
                "Circuit.call", "int/list[int]/Qubit/Qureg", type(indexes)
            )

        self._pointer = indexes
        return self

    def __getitem__(self, item):
        """ to fit the slice operator, overloaded this function.

        get a smaller qureg/qubit from this circuit

        Args:
            item(int/slice): slice passed in.
        Return:
            Qubit/Qureg: the result or slice
        """
        return self.qubits[item]

    def add_qubit(self, qubits: Union[Qureg, int], is_ancillary_qubit: bool = False):
        """ add additional qubits in circuit.

        Args:
            qubits Union[Qureg, int]: The new qubits.
            is_ancillae_qubit (bool, optional): whether the given qubits is ancillae, default to False.
        """
        if isinstance(qubits, int):
            if qubits <= 0:
                raise IndexExceedError("Circuit.add_qubit", ">= 0", {qubits})

            qubits = Qureg(qubits)

        self._qubits = self._qubits + qubits
        if is_ancillary_qubit:
            self._ancillae_qubits += list(range(self.width() - len(qubits), self.width()))

        logger.debug(f"Quantum Circuit {self._name} add {len(qubits)} qubits.")

    def reset_qubits(self):
        """ Reset all qubits in current circuit. """
        self._qubits.reset_qubits()
        logger.debug(f"Reset qubits' measured result in the Quantum Circuit {self._name}.")

    def remapping(self, qureg: Qureg, mapping: list, circuit_update: bool = False):
        """ Realignment the qubits by the given mapping.

        Args:
            qureg (Qureg): The qubits which need to permutate.
            mapping (list): The order of permutation.
            circuit_update (bool, optional): whether rearrange the qubits in circuit. Defaults to False.
        """
        if not isinstance(qureg, Qureg):
            raise TypeError("Circuit.remapping", "Qureg", type(qureg))

        if len(qureg) != len(mapping):
            raise ValueError("Circuit.remapping.mapping", len(qureg), len(mapping))

        current_index = [self.qubits.index(qubit) for qubit in qureg]
        remapping_index = [current_index[m] for m in mapping]
        remapping_qureg = self.qubits[remapping_index]

        if circuit_update:
            self._qubits = remapping_qureg
            logger.debug(f"The qureg is permutation by the order {mapping}.")

        qureg[:] = remapping_qureg

    ####################################################################
    ############          Circuit Gates Operators           ############
    ####################################################################
    def _add_quantumgate_into_circuit(self, gate: BasicGate, insert_idx: int = -1):
        if insert_idx == -1:
            self.gates.append(gate)
            insert_idx = len(self.gates)
        else:
            self.gates.insert(insert_idx, gate)

        logger.debug(
            f"Add quantum gate {gate.type} with qubit indexes {gate.cargs + gate.targs} " +
            f"with index {insert_idx}."
        )

        # Update gate type dict
        if gate.type in self._gate_type.keys():
            self._gate_type[gate.type] += 1
        else:
            self._gate_type[gate.type] = 1

    def _update_gate_index(self):
        for index, gate in enumerate(self.gates):
            gate_type, gate_qb, gate_idx = gate.name.split('-')

            if int(gate_idx) != index:
                gate.name = '-'.join([gate_type, gate_qb, str(index)])

    def replace_gate(self, idx: int, gate: BasicGate):
        """ Replace the quantum gate in the target index, only accept BasicGate or NoiseGate.

        Args:
            idx (int): The index of replaced quantum gate in circuit.
            gate (BasicGate): The new quantum gate
        """
        assert idx >= 0 and idx < len(self._gates), IndexExceedError(
            "Circuit.replace_gate.idx", [0, len(self._gates)], idx
        )
        assert isinstance(gate, (BasicGate, NoiseGate)), TypeError(
            "Circuit.replace_gate.gate", "[BasicGate, NoiseGate]", type(gate)
        )

        logger.debug(f"The origin gate {self._gates[idx]} is replaced by {gate}")
        self._gates[idx] = gate

    def find_position(self, cp_child: CheckPointChild):
        position = -1
        if cp_child is None:
            return position

        for cp in self._checkpoints:
            if cp.uid == cp_child.uid:
                position = cp.position
                cp.position = cp_child.shift
            elif position != -1:
                # change the related position for backward checkpoint
                cp.position = cp_child.shift

        return position

    def get_DAG_circuit(self) -> DAGCircuit:
        """
        Translate a quantum circuit to a directed acyclic graph
        via quantum gates dependencies (The commutation of quantum gates).

        The nodes in the graph represented the quantum gates, and the edges means the two quantum
        gates is non-commutation. In other words, a directed edge between node A with quantum gate GA
        and node B with quantum gate GB, the quantum gate GA does not commute with GB.

        The nodes in the graph have the following attributes:
        'name', 'gate', 'cargs', 'targs', 'qargs', 'successors', 'predecessors'.

        **Reference:**

        [1] Iten, R., Moyard, R., Metger, T., Sutter, D. and Woerner, S., 2020.
        Exact and practical pattern matching for quantum circuit optimization.
        `arXiv:1909.05270 <https://arxiv.org/abs/1909.05270>`_

        Returns:
            DAGCircuit: A directed acyclic graph represent current quantum circuit
        """
        try:
            return DAGCircuit(self)
        except Exception as e:
            raise CircuitDAGError(e)

    ####################################################################
    ############          Circuit Build Operators           ############
    ####################################################################
    def extend(self, gates: list):
        """ Add list of gates to the circuit

        Args:
            gates(list<BasicGate>): the gate to be added to the circuit
        """
        position = -1
        if not isinstance(gates, list):
            position = self.find_position(gates.checkpoint)
            gates = gates.gates

        for gate in gates:
            if position == -1:
                self.append(gate, is_extend=True)
            else:
                self.append(gate, is_extend=True, insert_idx=position)
                position += 1

        self._pointer = None

    def append(self, op: Union[BasicGate, Operator], is_extend: bool = False, insert_idx: int = -1):
        qureg = self._pointer[:] if self._pointer else None
        if not is_extend:
            self._pointer = None

        if isinstance(op, BasicGate):
            self._add_gate(op, qureg, insert_idx)
        elif isinstance(op, Trigger):
            self._add_trigger(op, qureg)
        elif isinstance(op, CheckPoint):
            self._checkpoints.append(op)
            logger.debug(f"Add an CheckPoint which point to index {op.position}.")
        elif isinstance(op, Operator):
            self._gates.append(op)
            logger.debug(f"Add an operator {type(op)}.")
        else:
            raise TypeError(
                "Circuit.append.gate", "Trigger/BasicGate/NoiseGate", {type(op)}
            )

    def _add_gate(self, gate: BasicGate, qureg: Qureg, insert_idx: int):
        """ add a gate into some qureg

        Args:
            gate(BasicGate)
            qureg(Qureg)
        """
        args_num = gate.controls + gate.targets
        gate_ctargs = gate.cargs + gate.targs
        is_assigned = gate.assigned_qubits or gate_ctargs
        if not qureg:
            if not is_assigned:
                if gate.is_single():
                    self._add_gate_to_all_qubits(gate)
                    return
                elif args_num == self.width():
                    qureg = self.qubits
                else:
                    raise CircuitAppendError(f"{gate.type} need assign qubits to add into circuit.")
            else:
                qureg = self.qubits[gate_ctargs] if gate_ctargs else gate.assigned_qubits
        else:
            if len(qureg) < args_num:
                raise CircuitAppendError("Assigned qubits must larger or equal to gate size.")

            if len(qureg) > args_num and gate_ctargs:
                qureg = qureg[gate_ctargs]

        gate = gate.copy()
        gate.cargs = [self.qubits.index(qureg[idx]) for idx in range(gate.controls)]
        gate.targs = [self.qubits.index(qureg[idx]) for idx in range(gate.controls, gate.controls + gate.targets)]
        gate.assigned_qubits = qureg
        gate.update_name(qureg[0].id, len(self.gates))

        # Add gate into circuit
        self._add_quantumgate_into_circuit(gate, insert_idx)

    def _add_gate_to_all_qubits(self, gate):
        for idx in range(self.width()):
            new_gate = gate.copy()
            new_gate.targs = [idx]
            new_gate.assigned_qubits = self.qubits(idx)
            new_gate.update_name(self.qubits[idx].id, len(self.gates))

            self._add_quantumgate_into_circuit(new_gate)

    def _add_trigger(self, op: Trigger, qureg: Qureg):
        if qureg:
            if len(qureg) != op.targets:
                raise CircuitAppendError("Failure to add Trigger into Circuit, as un-matched qureg.")

            op.targs = [self.qubits.index(qureg[idx]) for idx in range(op.targets)]
        else:
            if not op.targs:
                raise CircuitAppendError("Trigger need assign qubits to add into circuit.")

            for targ in op.targs:
                if targ >= self.width():
                    raise CircuitAppendError("The trigger's target exceed the width of the circuit.")

        self.gates.append(op)
        logger.debug(f"Add an operator Trigger with qubit indexes {op.targs}.")

    def random_append(
        self,
        rand_size: int = 10,
        typelist: list = None,
        random_params: bool = False,
        probabilities: list = None
    ):
        """ add some random gate to the circuit, not include Unitary, Permutation and Permutation_FX Gate.

        Args:
            rand_size(int): the number of the gate added to the circuit
            typelist(list<GateType>): the type of gate, default contains
                Rx, Ry, Rz, Cx, Cy, Cz, CRz, Ch, Rxx, Ryy, Rzz and FSim
            random_params(bool): whether using random parameters for all quantum gates with parameters.
            probabilities: The probability of append for each gates
        """
        if typelist is None:
            typelist = [
                GateType.rx, GateType.ry, GateType.rz,
                GateType.cx, GateType.cy, GateType.crz,
                GateType.ch, GateType.cz, GateType.rxx,
                GateType.ryy, GateType.rzz, GateType.fsim
            ]

        unsupported_gate_type = [GateType.unitary, GateType.perm, GateType.perm_fx]
        if len(set(typelist) & set(unsupported_gate_type)) != 0:
            raise CircuitSpecialAppendError(
                f"{set(typelist) & set(unsupported_gate_type)} is not support in random append."
            )

        if probabilities is not None:
            if not np.isclose(sum(probabilities), 1, atol=1e-6):
                raise ValueError("Circuit.random_append.probabilities", "sum to 1", sum(probabilities))

            if len(probabilities) != len(typelist):
                raise CircuitSpecialAppendError(
                    "The length of probabilities should equal to the length of Gate Typelist."
                )

        logger.debug(f"Random append {rand_size} quantum gates from {typelist} with probability {probabilities}.")
        gate_prob = probabilities
        gate_indexes = list(range(len(typelist)))
        n_qubit = self.width()
        for _ in range(rand_size):
            rand_type = np.random.choice(gate_indexes, p=gate_prob)
            gate_type = typelist[rand_type]
            self.append(build_random_gate(gate_type, n_qubit, random_params))

    def supremacy_append(self, repeat: int = 1, pattern: str = "ABCDCDAB"):
        """
        Add a supremacy circuit to the circuit

        Args:
            repeat(int): the number of two-qubit gates' sequence
            pattern(str): indicate the two-qubit gates' sequence
        """
        qubits = len(self.qubits)
        supremacy_layout = SupremacyLayout(qubits)
        supremacy_typelist = [GateType.sx, GateType.sy, GateType.sw]
        logger.debug(f"Append Supremacy Circuit with mapping pattern sequence {pattern} and repeat {repeat} times.")

        self._add_gate_to_all_qubits(H)

        for i in range(repeat * len(pattern)):
            for q in range(qubits):
                gate_type = supremacy_typelist[np.random.randint(0, 3)]
                self.append(build_gate(gate_type, q))

            current_pattern = pattern[i % (len(pattern))]
            if current_pattern not in "ABCD":
                raise ValueError("Circuit.append_supremacy.pattern", "[A, B, C, D]", current_pattern)

            edges = supremacy_layout.get_edges_by_pattern(current_pattern)
            for e in edges:
                gate_params = [np.pi / 2, np.pi / 6]
                gate_args = [int(e[0]), int(e[1])]
                fgate = build_gate(GateType.fsim, gate_args, gate_params)

                self.append(fgate)

        self._add_gate_to_all_qubits(Measure)

    ####################################################################
    ############                Circuit Utils               ############
    ####################################################################
    def sub_circuit(
        self,
        start: int = 0,
        max_size: int = -1,
        qubit_limit: Union[int, List[int], Qureg] = [],
        gate_limit: List[GateType] = [],
        remove: bool = False
    ):
        """ get a sub circuit

        Args:
            start(int): the start gate's index, default 0
            max_size(int): max size of the sub circuit, default -1 without limit
            qubit_limit(int/list<int>/Qureg): the required qubits' indexes, if [], accept all qubits. default to be [].
            gate_limit(List[GateType]): list of required gate's type, if [], accept all quantum gate. default to be [].
            remove(bool): whether deleting the slice gates from origin circuit, default False
        Return:
            Circuit: the sub circuit
        """
        max_size_for_logger = max_size if max_size == -1 else len(self.gates) - start
        logger.debug(
            f"Get {max_size_for_logger} gates from gate index {start}" +
            f" with target qubits {qubit_limit} and gate limit {gate_limit}."
        )
        if qubit_limit:
            target_qubits = qubit_limit
            if isinstance(qubit_limit, Qureg):
                target_qubits = [self._qubits.index(qubit) for qubit in qubit_limit]
            elif isinstance(qubit_limit, int):
                target_qubits = [qubit_limit]

            for target in target_qubits:
                if target < 0 or target >= self.width():
                    raise IndexExceedError("Circuit.sub_circuit.qubit_limit", [0, self.width()], target)

            set_tqubits = set(target_qubits)

        sub_circuit = Circuit(self.width()) if not qubit_limit else Circuit(len(target_qubits))
        sub_gates = self._gates[:]
        for gate_index in range(start, len(self._gates)):
            gate = sub_gates[gate_index]
            _gate = gate.copy()
            gate_args = set(gate.cargs + gate.targs)
            is_append_in_subc = True
            if (qubit_limit and gate_args & set(set_tqubits) != gate_args):
                is_append_in_subc = False

            if (gate_limit and gate.type not in gate_limit):
                is_append_in_subc = False

            if is_append_in_subc:
                if qubit_limit:
                    _gate.targs = [target_qubits.index(targ) for targ in _gate.targs]
                    _gate.cargs = [target_qubits.index(carg) for carg in _gate.cargs]
                    _gate | sub_circuit
                else:
                    _gate | sub_circuit

                if remove:
                    self._gates.remove(gate)

            if sub_circuit.size() >= max_size and max_size != -1:
                break

        if remove:
            self._update_gate_index()
            logger.warn(f"Remove sub-circuit's gates from quantum circuit {self._name}.")

        return sub_circuit

    def draw(self, method: str = 'matp_auto', filename: str = None):
        """Draw the figure of circuit.

        Args:
            method(str): the method to draw the circuit
                matp_inline: Show the figure interactively but do not save it to file.
                matp_file: Save the figure to file but do not show it interactively.
                matp_auto: Automatically select inline or file mode according to matplotlib backend.
                matp_silent: Return the drawn figure without saving or showing.
                command : command
            filename(str): the output filename without file extensions, default to None.
                If filename is None, it will using matlibplot.show() except matlibplot.backend
                is agg, it will output jpg file named circuit's name.
            get_figure(bool): Whether to return the figure object of matplotlib.

        Returns:
            If method is 'matp_silent', a matplotlib Figure is returned. Note that that figure is created in matplotlib
            Object Oriented interface, which means it must be display with IPython.display.

        Examples:
            >>> from IPython.display import display
            >>> circ = Circuit(5)
            >>> circ.random_append()
            >>> silent_fig = circ.draw(method="matp_silent")
            >>> display(silent_fig)
        """
        from QuICT.tools.drawer import PhotoDrawer, TextDrawing
        import matplotlib

        if method.startswith('matp'):
            if filename is not None:
                if '.' not in filename:
                    filename += '.jpg'

            photo_drawer = PhotoDrawer()
            if method == 'matp_auto':
                save_file = matplotlib.get_backend() == 'agg'
                show_inline = matplotlib.get_backend() != 'agg'
            elif method == 'matp_file':
                save_file = True
                show_inline = False
            elif method == 'matp_inline':
                save_file = False
                show_inline = True
            elif method == 'matp_silent':
                save_file = False
                show_inline = False
            else:
                raise ValueError(
                    "Circuit.draw.matp_method", "[matp_auto, matp_file, matp_inline, matp_silent]", method
                )

            silent = (not show_inline) and (not save_file)
            photo_drawer.run(circuit=self, filename=filename, save_file=save_file)

            if show_inline:
                from IPython.display import display
                display(photo_drawer.figure)
            elif silent:
                return photo_drawer.figure

        elif method == 'command':
            text_drawer = TextDrawing([i for i in range(len(self.qubits))], self.gates)
            if filename is None:
                print(text_drawer.single_string())
                return
            elif '.' not in filename:
                filename += '.txt'

            text_drawer.dump(filename)
        else:
            raise ValueError(
                "Circuit.draw.method", "[matp_auto, matp_file, matp_inline, matp_silent, command]", method
            )



from typing import Tuple
import numpy as np

from .data_structure import ResourceOp


def user_resource_op(user_info: dict, circuit_width: int, device: str, op: ResourceOp) -> Tuple[bool, dict]:
    signal = -1 if op == ResourceOp.Release else 1
    number_of_running_jobs = user_info['number_of_running_jobs'] + signal
    # exceed limit job number
    if number_of_running_jobs > user_info['maximum_parallel_level']:
        return False, user_info

    # Deal with running job's resource
    if device == "CPU":
        cur_user_running_qubits = _qubits_operator(
            user_info['running_qubits_in_cpu'], circuit_width, signal
        )
        qubits_limitation = user_info['max_qubits_in_cpu']
    else:
        cur_user_running_qubits = _qubits_operator(
            user_info['running_qubits_in_gpu'], circuit_width, signal
        )
        qubits_limitation = user_info['max_qubits_in_gpu']

    # exceed limit qubits number
    if cur_user_running_qubits > qubits_limitation:
        return False, user_info

    # Update user's infomation
    user_info['number_of_running_jobs'] = number_of_running_jobs
    if device == "CPU":
        user_info['running_qubits_in_cpu'] = cur_user_running_qubits
    else:
        user_info['running_qubits_in_gpu'] = cur_user_running_qubits

    return True, user_info


def _qubits_operator(qa, qb: int, signal):
    assert qb > 0
    if signal == -1:
        assert np.isclose(qa, qb) or qa > qb

    if qa == 0:
        return qb

    result = 2 ** qa + signal * 2 ** qb
    return np.log2(result) if not np.isclose(result, 0) else 0



from typing import *
import numpy as np

from scipy.linalg import cossin
from QuICT.core import *
from QuICT.core.gate import *
from QuICT.core.gate.backend import UniformlyRotation

from .cartan_kak_decomposition import CartanKAKDecomposition
from .cartan_kak_diagonal_decomposition import CartanKAKDiagonalDecomposition
from .uniformly_ry_revision import UniformlyRyRevision
from .utility import *


class UnitaryDecomposition(object):
    def __init__(self, include_phase_gate: bool = True, recursive_basis: int = 2):
        """
        Args:
            include_phase_gate(bool): Whether to include a phase gate to keep synthesized gate matrix the same
                as input. If set False, the output gates might have a matrix which has a factor shift to input:
                np.allclose(<matrix_of_return_gates> * factor, <input_matrix>).
            recursive_basis(int): Terminate recursion at which level. It could be set as 1 or 2, which would stop
                recursion when matrix is 2 or 4, respectively. When set as 2, the final step is done by KAK
                decomposition. Correctness of this algorithm is never influenced by recursive_basis.
        """
        self.include_phase_gate = include_phase_gate
        self.recursive_basis = recursive_basis

    def execute(self, mat: np.ndarray):
        """
        Transform a general unitary matrix into CX gates and single qubit gates.

        Args:
            mat(np.ndarray): Unitary matrix.

        Returns:
            Union[Tuple[CompositeGate,None], Tuple[CompositeGate,complex]]: If inlclude_phase_gate==False,
            this function returns synthesized gates and a shift factor. Otherwise a tuple like (<gates>, None)
            is returned.
        """
        qubit_num = int(round(np.log2(mat.shape[0])))

        """
        After adding KAK diagonal optimization, inner built gates would have a
        leading diagonal gate not decomposed. If our first-level recursion is
        2-bit, using older version of KAK is OK.
        """

        if qubit_num == 2 and self.recursive_basis == 2:
            CKD = CartanKAKDecomposition()
            gates = CKD.execute(mat)
            syn_mat = gates.matrix()
            shift = shift_ratio(mat, syn_mat)
            add_factor_shift_into_phase(gates, shift)
            return gates, None

        gates, shift = self.inner_utrans_build_gate(
            mat=mat,
            recursive_basis=self.recursive_basis,
            keep_left_diagonal=False,
        )

        if self.recursive_basis == 2 and self.include_phase_gate:
            gates = add_factor_shift_into_phase(gates, shift)
        if self.include_phase_gate:
            return gates, None
        else:
            return gates, shift

    def inner_utrans_build_gate(
        self,
        mat: np.ndarray,
        recursive_basis: int = 1,
        keep_left_diagonal: bool = False,
        use_cz_ry: bool = True,
    ) -> Tuple[CompositeGate, complex]:
        mat: np.ndarray = np.array(mat)
        mat_size: int = mat.shape[0]
        qubit_num = int(round(np.log2(mat_size)))

        _kak = CartanKAKDiagonalDecomposition() if keep_left_diagonal else CartanKAKDecomposition()

        if qubit_num == 1:
            u = build_gate(GateType.unitary, [0], mat)
            _ret = CompositeGate(gates=[u])
            return _ret, 1.0 + 0.0j
        elif qubit_num == 2 and recursive_basis == 2:
            gates = _kak.execute(mat)
            syn_mat = gates.matrix()
            shift = shift_ratio(mat, syn_mat)
            return gates, shift

        u, angle_list, v_dagger = cossin(mat, mat_size // 2, mat_size // 2, separate=True)

        """
        Parts of following comments are from Scipy documentation
        (https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.cossin.html)
                                   ┌                   ┐
                                   │ I  0  0 │ 0  0  0 │
        ┌           ┐   ┌         ┐│ 0  C  0 │ 0 -S  0 │┌         ┐*
        │ X11 │ X12 │   │ U1 │    ││ 0  0  0 │ 0  0 -I ││ V1 │    │
        │ ────┼──── │ = │────┼────││─────────┼─────────││────┼────│
        │ X21 │ X22 │   │    │ U2 ││ 0  0  0 │ I  0  0 ││    │ V2 │
        └           ┘   └         ┘│ 0  S  0 │ 0  C  0 │└         ┘
                                   │ 0  0  I │ 0  0  0 │
                                   └                   ┘

        Both u and v are controlled unitary operations hence can be
        decomposed into 2 (smaller) unitary operations and 1 controlled rotation.
        """

        # Dynamically import to avoid circulation.
        from .controlled_unitary import ControlledUnitaryDecomposition

        shift: complex = 1.0

        # (c,s\\s,c)
        angle_list *= 2  # Ry use its angle as theta/2
        if use_cz_ry:
            URyRevision = UniformlyRyRevision(is_cz_left=False)  # keep CZ at right side
            reversed_ry = URyRevision.execute(angle_list)
            reversed_ry & [(i + 1) % qubit_num for i in range(qubit_num)]
        else:
            URy = UniformlyRotation(GateType.ry)
            reversed_ry = URy.execute(angle_list)
            reversed_ry & [(i + 1) % qubit_num for i in range(qubit_num)]

        """
        Now, gates have CZ gate(s) at it's ending part(the left side of u).
        Left gate of u is right multiplied to u in matrix view.
        If qubit_num > 2, we would have reversed_ry[-1] as a CZ affecting on (0, 1),
        while reversed_ry[-2] a CZ on (0, qubit_num - 1).
        If qubit_num == 2, there would only be one CZ affecting on (0, 1).
        """

        # u
        u1: np.ndarray = u[0]
        u2: np.ndarray = u[1]

        if use_cz_ry:
            reversed_ry.gates.pop()  # CZ on (0,1)
            # This CZ affects 1/4 last columns of the matrix of U, or 1/2 last columns of u2.
            _u_size = u2.shape[0]
            for i in range(_u_size // 2, _u_size):
                u2[:, i] = -u2[:, i]

            if qubit_num > 2:
                reversed_ry.gates.pop()  # CZ on (0, qubit_num - 1)
                # For similar reasons, this CZ only affect 2 parts of matrix of U.
                for i in range(_u_size - _u_size // 4, _u_size):
                    u1[:, i] = - u1[:, i]
                    u2[:, i] = - u2[:, i]

        CUD = ControlledUnitaryDecomposition()
        u_gates, _shift = CUD.inner_cutrans_build_gate(
            u1=u1,
            u2=u2,
            recursive_basis=recursive_basis,
            keep_left_diagonal=True,
        )
        shift *= _shift

        # v_dagger

        """
        Now, leftmost gate decompose by u is a diagonal gate on (n-2, n-1), which
        is commutable with reversed_ry. That gate is in right side of v_dagger, so
        it would be left multiplied to v_dagger.
        """
        v1_dagger = v_dagger[0]
        v2_dagger = v_dagger[1]

        if recursive_basis == 2:
            forwarded_d_gate: BasicGate = u_gates.gates.pop(0)
            forwarded_mat = forwarded_d_gate.matrix
            for i in range(0, mat_size // 2, 4):
                for k in range(4):
                    v1_dagger[i + k, :] *= forwarded_mat[k, k]
                    v2_dagger[i + k, :] *= forwarded_mat[k, k]

        v_dagger_gates, _shift = CUD.inner_cutrans_build_gate(
            u1=v1_dagger,
            u2=v2_dagger,
            recursive_basis=recursive_basis,
            keep_left_diagonal=keep_left_diagonal,
        )
        shift *= _shift

        gates = CompositeGate()
        gates.extend(v_dagger_gates)
        gates.extend(reversed_ry)
        gates.extend(u_gates)

        return gates, shift
